@@ -54,13 +54,6 @@
 #define SHADOW_COLOR_GREEN 0x0000
 #define SHADOW_COLOR_BLUE  0x0000
 
-#define DBUS_DEST           "org.freedesktop.compiz"
-#define DBUS_SIGNAL_PATH    "/org/freedesktop/compiz/decoration/display"
-#define DBUS_QUERY_PATH     "/org/freedesktop/compiz/decoration/allscreens"
-#define DBUS_INTERFACE      "org.freedesktop.compiz"
-#define DBUS_METHOD_GET     "get"
-#define DBUS_SIGNAL_CHANGED "changed"
-
 int    blurType = BLUR_TYPE_NONE;
 
 decor_shadow_t *KWD::Decorator::mNoBorderShadow = 0;
@@ -68,7 +61,12 @@ KWD::PluginManager *KWD::Decorator::mPlugins = 0;
 KWD::Options *KWD::Decorator::mOptions = 0;
 NETRootInfo *KWD::Decorator::mRootInfo;
 WId KWD::Decorator::mActiveId;
-decor_shadow_options_t KWD::Decorator::mShadowOptions;
+decor_shadow_options_t KWD::Decorator::mActiveShadowOptions;
+decor_shadow_options_t KWD::Decorator::mInactiveShadowOptions;
+
+KWD::Window *KWD::Decorator::mDecorNormal = NULL;
+KWD::Window *KWD::Decorator::mDecorActive = NULL;
+KWD::Decorator *KWD::Decorator::mSelf = NULL;
 
 struct _cursor cursors[3][3] = {
     { C (top_left_corner), C (top_side), C (top_right_corner) },
@@ -94,7 +92,8 @@ KWD::Decorator::Decorator () :
 {
     XSetWindowAttributes attr;
     int			 i, j;
-    QDBusConnection      dbus = QDBusConnection::sessionBus ();
+
+    mSelf = this;
 
     mRootInfo = new NETRootInfo (QX11Info::display (), 0);
 
@@ -107,42 +106,29 @@ KWD::Decorator::Decorator () :
     Atoms::init ();
 
     (void *) new KWinAdaptor (this);
-    dbus.registerObject ("/KWin", this);
-    dbus.connect (QString (), "/KWin", "org.kde.KWin", "reloadConfig", this,
-		  SLOT (reconfigure ()));
-
-    dbus.connect (QString (), DBUS_SIGNAL_PATH "/shadow_radius",
-		  DBUS_INTERFACE, DBUS_SIGNAL_CHANGED, this,
-		  SLOT (shadowRadiusChanged (double)));
-
-    dbus.connect (QString (), DBUS_SIGNAL_PATH "/shadow_opacity",
-		  DBUS_INTERFACE, DBUS_SIGNAL_CHANGED, this,
-		  SLOT (shadowOpacityChanged (double)));
-
-    dbus.connect (QString (), DBUS_SIGNAL_PATH "/shadow_x_offset",
-		  DBUS_INTERFACE, DBUS_SIGNAL_CHANGED, this,
-		  SLOT (shadowXOffsetChanged (int)));
-
-    dbus.connect (QString (), DBUS_SIGNAL_PATH "/shadow_y_offset",
-		  DBUS_INTERFACE, DBUS_SIGNAL_CHANGED, this,
-		  SLOT (shadowYOffsetChanged (int)));
-
-    dbus.connect (QString (), DBUS_SIGNAL_PATH "/shadow_color",
-		  DBUS_INTERFACE, DBUS_SIGNAL_CHANGED, this,
-		  SLOT (shadowColorChanged (QString)));
 
     mConfig = new KConfig ("kwinrc");
 
     mOptions = new KWD::Options (mConfig);
     mPlugins = new PluginManager (KSharedConfig::openConfig ("kwinrc"));
 
-    mShadowOptions.shadow_radius   = SHADOW_RADIUS;
-    mShadowOptions.shadow_opacity  = SHADOW_OPACITY;
-    mShadowOptions.shadow_offset_x = SHADOW_OFFSET_X;
-    mShadowOptions.shadow_offset_y = SHADOW_OFFSET_Y;
-    mShadowOptions.shadow_color[0] = SHADOW_COLOR_RED;
-    mShadowOptions.shadow_color[1] = SHADOW_COLOR_GREEN;
-    mShadowOptions.shadow_color[2] = SHADOW_COLOR_BLUE;
+    mActiveShadowOptions.shadow_radius   = SHADOW_RADIUS;
+    mActiveShadowOptions.shadow_opacity  = SHADOW_OPACITY;
+    mActiveShadowOptions.shadow_offset_x = SHADOW_OFFSET_X;
+    mActiveShadowOptions.shadow_offset_y = SHADOW_OFFSET_Y;
+    mActiveShadowOptions.shadow_color[0] = SHADOW_COLOR_RED;
+    mActiveShadowOptions.shadow_color[1] = SHADOW_COLOR_GREEN;
+    mActiveShadowOptions.shadow_color[2] = SHADOW_COLOR_BLUE;
+
+    mInactiveShadowOptions.shadow_radius   = SHADOW_RADIUS;
+    mInactiveShadowOptions.shadow_opacity  = SHADOW_OPACITY;
+    mInactiveShadowOptions.shadow_offset_x = SHADOW_OFFSET_X;
+    mInactiveShadowOptions.shadow_offset_y = SHADOW_OFFSET_Y;
+    mInactiveShadowOptions.shadow_color[0] = SHADOW_COLOR_RED;
+    mInactiveShadowOptions.shadow_color[1] = SHADOW_COLOR_GREEN;
+    mInactiveShadowOptions.shadow_color[2] = SHADOW_COLOR_BLUE;
+
+    updateShadowProperties (QX11Info::appRootWindow ());
 
     for (i = 0; i < 3; i++)
     {
@@ -164,7 +150,7 @@ KWD::Decorator::Decorator () :
 				      CopyFromParent,
 				      CopyFromParent,
 				      CWOverrideRedirect, &attr);
-				      
+
     long data = 1;
     XChangeProperty (QX11Info::display(), mCompositeWindow, Atoms::enlightmentDesktop,
 		      XA_CARDINAL, 32, PropModeReplace, (unsigned char *) &data, 1);
@@ -203,6 +189,10 @@ bool
 KWD::Decorator::enableDecorations (Time timestamp)
 {
     QList <WId>::ConstIterator it;
+    unsigned int nchildren;
+    WId       *children;
+    WId       root, parent;
+    long int  select;
 
     mDmSnTimestamp = timestamp;
 
@@ -217,9 +207,10 @@ KWD::Decorator::enableDecorations (Time timestamp)
 
     updateShadow ();
 
+    /* FIXME: Implement proper decoration lists and remove this */
     mDecorNormal = new KWD::Window (mCompositeWindow,
-				    QX11Info::appRootWindow (),
-				    0, Window::Default);
+                                    QX11Info::appRootWindow (),
+                                    0, Window::Default);
     mDecorActive = new KWD::Window (mCompositeWindow,
 				    QX11Info::appRootWindow (),
 				    0, Window::DefaultActive);
@@ -239,12 +230,33 @@ KWD::Decorator::enableDecorations (Time timestamp)
     foreach (WId id, KWindowSystem::windows ())
 	handleWindowAdded (id);
 
+    /* Find the switcher and add it too
+     * FIXME: Doing XQueryTree and then
+     * XGetWindowProperty on every window
+     * like this is really expensive, surely
+     * there is a better way to do this */
+
+    XQueryTree (QX11Info::display (), QX11Info::appRootWindow (),
+                &root, &parent, &children, &nchildren);
+
+    for (unsigned int i = 0; i < nchildren; i++)
+    {
+        if (KWD::readWindowProperty (children[i],
+                                     Atoms::switchSelectWindow, &select))
+        {
+            handleWindowAdded(children[i]);
+            break;
+        }
+    }
+
     connect (Plasma::Theme::defaultTheme (), SIGNAL (themeChanged ()),
 	     SLOT (plasmaThemeChanged ()));
 
     // select for client messages
     XSelectInput (QX11Info::display (), QX11Info::appRootWindow (),
-		  StructureNotifyMask | PropertyChangeMask);
+                  SubstructureNotifyMask |
+                  StructureNotifyMask |
+                  PropertyChangeMask);
 
     return true;
 }
@@ -252,72 +264,28 @@ KWD::Decorator::enableDecorations (Time timestamp)
 void
 KWD::Decorator::updateAllShadowOptions (void)
 {
-    QDBusInterface       *compiz;
-    QDBusReply<QString>  stringReply;
-    QDBusReply<double>   doubleReply;
-    QDBusReply<int>      intReply;
-    int                  c[4];
-
-    compiz = new QDBusInterface (DBUS_DEST, DBUS_QUERY_PATH "/shadow_radius",
-				 DBUS_INTERFACE);
-    doubleReply = compiz->call (DBUS_METHOD_GET);
-    delete compiz;
-
-    if (doubleReply.isValid ())
-	mShadowOptions.shadow_radius = doubleReply.value ();
-
-    compiz = new QDBusInterface (DBUS_DEST, DBUS_QUERY_PATH "/shadow_opacity",
-				 DBUS_INTERFACE);
-    doubleReply = compiz->call (DBUS_METHOD_GET);
-    delete compiz;
-
-    if (doubleReply.isValid ())
-	mShadowOptions.shadow_opacity = doubleReply.value ();
-
-    compiz = new QDBusInterface (DBUS_DEST, DBUS_QUERY_PATH "/shadow_x_offset",
-				 DBUS_INTERFACE);
-    intReply = compiz->call (DBUS_METHOD_GET);
-    delete compiz;
-
-    if (intReply.isValid ())
-	mShadowOptions.shadow_offset_x = intReply.value ();
-
-    compiz = new QDBusInterface (DBUS_DEST, DBUS_QUERY_PATH "/shadow_y_offset",
-				 DBUS_INTERFACE);
-    intReply = compiz->call (DBUS_METHOD_GET);
-    delete compiz;
-
-    if (intReply.isValid ())
-	mShadowOptions.shadow_offset_y = intReply.value ();
-    else
-	mShadowOptions.shadow_offset_y = SHADOW_OFFSET_Y;
-
-    compiz = new QDBusInterface (DBUS_DEST, DBUS_QUERY_PATH "/shadow_color",
-				 DBUS_INTERFACE);
-    stringReply = compiz->call (DBUS_METHOD_GET);
-    delete compiz;
-
-    if (stringReply.isValid () &&
-	sscanf (stringReply.value ().toAscii ().data (), "#%2x%2x%2x%2x",
-		&c[0], &c[1], &c[2], &c[3]) == 4)
-    {
-	mShadowOptions.shadow_color[0] = c[0] << 8 | c[0];
-	mShadowOptions.shadow_color[1] = c[1] << 8 | c[1];
-	mShadowOptions.shadow_color[2] = c[2] << 8 | c[2];
-    }
+    updateShadowProperties (QX11Info::appRootWindow ());
 }
 
 void
-KWD::Decorator::changeShadowOptions (decor_shadow_options_t *opt)
+KWD::Decorator::changeShadowOptions (decor_shadow_options_t *aopt, decor_shadow_options_t *iopt)
 {
-    QMap <WId, KWD::Window *>::ConstIterator it;
+    bool changed = false;
 
-    if (!memcmp (opt, &mShadowOptions, sizeof (decor_shadow_options_t)))
-	return;
+    if (memcmp (aopt, &mActiveShadowOptions, sizeof (decor_shadow_options_t)))
+    {
+	mActiveShadowOptions = *aopt;
+	changed = true;
+    }
 
-    mShadowOptions = *opt;
+    if (memcmp (aopt, &mInactiveShadowOptions, sizeof (decor_shadow_options_t)))
+    {
+	mInactiveShadowOptions = *iopt;
+	changed = true;
+    }
 
-    updateShadow ();
+    if (changed)
+	updateShadow ();
 }
 
 void
@@ -340,7 +308,7 @@ KWD::Decorator::updateShadow (void)
 					   0,
 					   0,
 					   0, 0, 0, 0,
-					   &mShadowOptions,
+					   &mActiveShadowOptions,
 					   &context,
 					   decor_draw_simple,
 					   0);
@@ -348,7 +316,8 @@ KWD::Decorator::updateShadow (void)
     if (mNoBorderShadow)
     {
 	decor_extents_t extents = { 0, 0, 0, 0 };
-	long	        data[256];
+	long	        *data;
+	unsigned int    n = 1, frame_type = 0, frame_state = 0, frame_actions = 0;
 	decor_quad_t    quads[N_QUADS_MAX];
 	int	        nQuad;
 	decor_layout_t  layout;
@@ -357,20 +326,85 @@ KWD::Decorator::updateShadow (void)
 
 	nQuad = decor_set_lSrStSbS_window_quads (quads, &context, &layout);
 
-	decor_quads_to_property (data, mNoBorderShadow->pixmap,
-				 &extents, &extents,
-				 0, 0, quads, nQuad);
+	data = decor_alloc_property (n, WINDOW_DECORATION_TYPE_PIXMAP);
+	decor_quads_to_property (data, n - 1, mNoBorderShadow->pixmap,
+				 &extents, &extents, &extents, &extents,
+				 0, 0, quads, nQuad, frame_type, frame_state, frame_actions);
 
 	KWD::trapXError ();
 	XChangeProperty (QX11Info::display (), QX11Info::appRootWindow (),
 			 Atoms::netWindowDecorBare,
 			 XA_INTEGER,
 			 32, PropModeReplace, (unsigned char *) data,
-			 BASE_PROP_SIZE + QUAD_PROP_SIZE * nQuad);
+			 PROP_HEADER_SIZE + BASE_PROP_SIZE + QUAD_PROP_SIZE * N_QUADS_MAX);
 	KWD::popXError ();
+
+        free (data);
     }
 }
 
+void
+KWD::Decorator::updateShadowProperties (WId id)
+{
+    int nItems;
+    long *data;
+    double aradius, aopacity;
+    int    axOffset, ayOffset;
+    double iradius, iopacity;
+    int    ixOffset, iyOffset;
+    QVector<QString> shadowColor;
+
+    if (id != QX11Info::appRootWindow ())
+	return;
+
+    void *propData = KWD::readXProperty (id,
+					  Atoms::compizShadowInfo,
+					  XA_INTEGER,
+					  &nItems);
+
+    if (nItems != 4)
+	return;
+
+    data = reinterpret_cast <long *> (propData);
+
+    aradius = data[0];
+    aopacity = data[1];
+
+    /* We multiplied by 1000 in compiz to keep
+      * precision, now divide by that much */
+
+    aradius /= 1000;
+    aopacity /= 1000;
+
+    axOffset = data[2];
+    ayOffset = data[3];
+
+    iradius = data[4];
+    iopacity = data[5];
+
+    /* We multiplied by 1000 in compiz to keep
+     * precision, now divide by that much */
+
+    iradius /= 1000;
+    iopacity /= 1000;
+
+    ixOffset = data[6];
+    iyOffset = data[7];
+
+
+    shadowRadiusChanged (aradius, iradius);
+    shadowOpacityChanged (aopacity, iopacity);
+    shadowXOffsetChanged (axOffset, ixOffset);
+    shadowYOffsetChanged (ayOffset, iyOffset);
+
+    shadowColor = KWD::readPropertyString (id, Atoms::compizShadowColor);
+
+    if (shadowColor.size () == 2)
+	shadowColorChanged (shadowColor.at (0), shadowColor.at (1));
+
+    XFree (propData);
+}
+
 bool
 KWD::Decorator::x11EventFilter (XEvent *xevent)
 {
@@ -395,7 +429,24 @@ KWD::Decorator::x11EventFilter (XEvent *xevent)
 	if (status == DECOR_SELECTION_GIVE_UP)
 	    KApplication::exit (0);
 
-	break;
+        break;
+    case CreateNotify:
+        /* We only care about windows that aren't managed here */
+        if (!KWindowSystem::hasWId (xevent->xcreatewindow.window))
+        {
+            WId select;
+
+            KWD::trapXError ();
+            XSelectInput (QX11Info::display (), xevent->xcreatewindow.window,
+                          StructureNotifyMask | PropertyChangeMask);
+            KWD::popXError ();
+
+            if (KWD::readWindowProperty (xevent->xcreatewindow.window,
+                                         Atoms::switchSelectWindow,
+                                         (long *) &select))
+                handleWindowAdded (xevent->xcreatewindow.window);
+        }
+
     case PropertyNotify:
 	if (xevent->xproperty.atom == Atoms::netInputFrameWindow)
 	{
@@ -405,6 +456,11 @@ KWD::Decorator::x11EventFilter (XEvent *xevent)
 	{
 	    handleWindowAdded (xevent->xproperty.window);
 	}
+	else if (xevent->xproperty.atom == Atoms::compizShadowInfo ||
+		 xevent->xproperty.atom == Atoms::compizShadowColor)
+	{
+	    updateShadowProperties (xevent->xproperty.window);
+	}
 	else if (xevent->xproperty.atom == Atoms::switchSelectWindow)
 	{
 	    WId id = xevent->xproperty.window;
@@ -428,7 +484,7 @@ KWD::Decorator::x11EventFilter (XEvent *xevent)
 	    break;
 
 	client = mFrames[xce->window];
-	
+
 	if (!client->decorWidget ())
 	    break;
 
@@ -468,7 +524,7 @@ KWD::Decorator::x11EventFilter (XEvent *xevent)
 	    break;
 
 	client = mFrames[xme->window];
-	
+
 	if (!client->decorWidget ())
 	    break;
 
@@ -512,7 +568,7 @@ KWD::Decorator::x11EventFilter (XEvent *xevent)
 	    break;
 
 	client = mFrames[xbe->window];
-	
+
 	if (!client->decorWidget ())
 	    break;
 
@@ -523,11 +579,11 @@ KWD::Decorator::x11EventFilter (XEvent *xevent)
 	    XButtonEvent xbe2 = *xbe;
 	    xbe2.window = child->winId ();
 	    QPoint p;
-		
+
 	    p = client->mapToChildAt (QPoint (xbe->x, xbe->y));
 	    xbe2.x = p.x ();
 	    xbe2.y = p.y ();
-	    
+
 	    p = child->mapToGlobal(p);
 	    xbe2.x_root = p.x ();
 	    xbe2.y_root = p.y ();
@@ -629,6 +685,7 @@ KWD::Decorator::handleWindowAdded (WId id)
     WId					     oframe = 0, iframe = 0;
     KWD::Window::Type			     type = KWD::Window::Normal;
     QWidgetList				     widgets;
+    QRect                                    geometry;
 
     /* avoid adding any of our own top level windows */
     foreach (QWidget *widget, QApplication::topLevelWidgets ()) {
@@ -636,19 +693,6 @@ KWD::Decorator::handleWindowAdded (WId id)
 	    return;
     }
 
-    KWD::trapXError ();
-    KWindowInfo wInfo = KWindowSystem::windowInfo (id, NET::WMGeometry);
-    if (KWD::popXError ())
-	return;
-
-    if (!wInfo.valid ())
-	return;
-
-    QRect geometry = wInfo.geometry ();
-
-    KWD::readWindowProperty (id, Atoms::netInputFrameWindow, (long *) &iframe);
-    KWD::readWindowProperty (id, Atoms::netOutputFrameWindow, (long *) &oframe);
-
     if (KWD::readWindowProperty (id, Atoms::switchSelectWindow,
 				 (long *) &select))
     {
@@ -659,11 +703,28 @@ KWD::Decorator::handleWindowAdded (WId id)
             delete mSwitcher;
             mSwitcher = new Switcher (mCompositeWindow, id);
         }
+
+	geometry = mSwitcher->geometry ();
 	frame = None;
     }
     else
     {
-	KWindowInfo wInfo = KWindowSystem::windowInfo (id, NET::WMWindowType, 0);
+        KWindowInfo wInfo;
+
+        KWD::trapXError ();
+        wInfo = KWindowSystem::windowInfo (id, NET::WMGeometry);
+        if (KWD::popXError ())
+            return;
+
+        if (!wInfo.valid ())
+            return;
+
+        KWD::readWindowProperty (id, Atoms::netInputFrameWindow, (long *) &iframe);
+        KWD::readWindowProperty (id, Atoms::netOutputFrameWindow, (long *) &oframe);
+
+        geometry = wInfo.geometry ();
+
+        wInfo = KWindowSystem::windowInfo (id, NET::WMWindowType, 0);
 
 	switch (wInfo.windowType (~0)) {
 	case NET::Normal:
@@ -794,7 +855,8 @@ KWD::Decorator::handleWindowChanged (WId		 id,
 
     if (mSwitcher && mSwitcher->xid () == id)
     {
-	mSwitcher->updateGeometry ();
+	if (properties[0] & NET::WMGeometry)
+	    mSwitcher->updateGeometry ();
 	return;
     }
 
@@ -849,61 +911,78 @@ KWD::Decorator::sendClientMessage (WId  eventWid,
 }
 
 void
-KWD::Decorator::shadowRadiusChanged (double value)
+KWD::Decorator::shadowRadiusChanged (double value_active, double value_inactive)
 {
-    decor_shadow_options_t opt = *shadowOptions ();
+    decor_shadow_options_t aopt = *activeShadowOptions ();
+    decor_shadow_options_t iopt = *inactiveShadowOptions ();
 
-    opt.shadow_radius = value;
+    aopt.shadow_radius = value_active;
+    iopt.shadow_radius = value_inactive;
 
-    changeShadowOptions (&opt);
+    changeShadowOptions (&aopt, &iopt);
 }
 
 void
-KWD::Decorator::shadowOpacityChanged (double value)
+KWD::Decorator::shadowOpacityChanged (double value_active, double value_inactive)
 {
-    decor_shadow_options_t opt = *shadowOptions ();
+    decor_shadow_options_t aopt = *activeShadowOptions ();
+    decor_shadow_options_t iopt = *inactiveShadowOptions ();
 
-    opt.shadow_opacity = value;
+    aopt.shadow_opacity = value_active;
+    iopt.shadow_opacity = value_inactive;
 
-    changeShadowOptions (&opt);
+    changeShadowOptions (&aopt, &iopt);
 }
 
 void
-KWD::Decorator::shadowXOffsetChanged (int value)
+KWD::Decorator::shadowXOffsetChanged (int value_active, int value_inactive)
 {
-    decor_shadow_options_t opt = *shadowOptions ();
+    decor_shadow_options_t aopt = *activeShadowOptions ();
+    decor_shadow_options_t iopt = *inactiveShadowOptions ();
 
-    opt.shadow_offset_x = value;
+    aopt.shadow_offset_x = value_active;
+    iopt.shadow_offset_x = value_inactive;
 
-    changeShadowOptions (&opt);
+    changeShadowOptions (&aopt, &iopt);
 }
 
 void
-KWD::Decorator::shadowYOffsetChanged (int value)
+KWD::Decorator::shadowYOffsetChanged (int value_active, double value_inactive)
 {
-    decor_shadow_options_t opt = *shadowOptions ();
+    decor_shadow_options_t aopt = *activeShadowOptions ();
+    decor_shadow_options_t iopt = *inactiveShadowOptions ();
 
-    opt.shadow_offset_y = value;
+    aopt.shadow_offset_y = value_active;
+    iopt.shadow_offset_y = value_inactive;
 
-    changeShadowOptions (&opt);
+    changeShadowOptions (&aopt, &iopt);
 }
 
 void
-KWD::Decorator::shadowColorChanged (QString value)
+KWD::Decorator::shadowColorChanged (QString value_active, QString value_inactive)
 {
-    decor_shadow_options_t opt = *shadowOptions ();
+    decor_shadow_options_t aopt = *activeShadowOptions ();
+    decor_shadow_options_t iopt = *inactiveShadowOptions ();
 
     int c[4];
 
-    if (sscanf (value.toAscii ().data (), "#%2x%2x%2x%2x",
+    if (sscanf (value_active.toAscii ().data (), "#%2x%2x%2x%2x",
 	        &c[0], &c[1], &c[2], &c[3]) == 4)
     {
-	opt.shadow_color[0] = c[0] << 8 | c[0];
-	opt.shadow_color[1] = c[1] << 8 | c[1];
-	opt.shadow_color[2] = c[2] << 8 | c[2];
+	aopt.shadow_color[0] = c[0] << 8 | c[0];
+	aopt.shadow_color[1] = c[1] << 8 | c[1];
+	aopt.shadow_color[2] = c[2] << 8 | c[2];
     }
 
-    changeShadowOptions (&opt);
+    if (sscanf (value_inactive.toAscii ().data (), "#%2x%2x%2x%2x",
+		&c[0], &c[1], &c[2], &c[3]) == 4)
+    {
+	iopt.shadow_color[0] = c[0] << 8 | c[0];
+	iopt.shadow_color[1] = c[1] << 8 | c[1];
+	iopt.shadow_color[2] = c[2] << 8 | c[2];
+    }
+
+    changeShadowOptions (&aopt, &iopt);
 }
 
 void
