@@ -1,3 +1,4 @@
+
 /*
  * Copyright © 2006 Novell, Inc.
  *
@@ -46,6 +47,8 @@ decor_version (void)
 
   data[1] = decoration type
 
+  data[2] = number of decorations specified in property
+
   WINDOW_DECORATION_TYPE_WINDOW property
   --------------------------------------
   data[3] = input left
@@ -61,24 +64,54 @@ decor_version (void)
   data[11] = min width
   data[12] = min height
 
+  fields 13 to 15 are only used by the default
+  decorations on the root window
+
+  data[13] = frame state
+  data[14] = frame type
+  data[15] = frame actions
+
 
   WINDOW_DECORATION_TYPE_PIXMAP property
   --------------------------------------
 
-  data[2] = pixmap
+  data[3] = pixmap
 
-  data[3] = input left
-  data[4] = input right
-  data[5] = input top
-  data[6] = input bottom
+  extents
 
-  data[7]  = input left when maximized
-  data[8]  = input right when maximized
-  data[9]  = input top when maximized
-  data[10] = input bottom when maximized
+  frame input is used for creating the input area of
+  the frame window which the client will be
+  reparented into, border is used for positioning
 
-  data[11] = min width
-  data[12] = min height
+  data[4] = frame left
+  data[5] = frame right
+  data[6] = frame top
+  data[7] = frame bottom
+  data[8] = input left
+  data[9] = input right
+  data[10] = input top
+  data[11] = input bottom
+
+  data[12]  = frame left when maximized
+  data[13]  = frame right when maximized
+  data[14]  = frame top when maximized
+  data[15] = frame bottom when maximized
+  data[16]  = border left when maximized
+  data[17]  = border right when maximized
+  data[18]  = border top when maximized
+  data[19] = border bottom when maximized
+
+  data[20] = min width
+  data[21] = min height
+
+  fields 22 to 24 are only used by the default
+  decorations on the root window
+
+  data[22] = frame state
+  data[23] = frame type
+  data[24] = frame actions
+
+  data[25] = num quads
 
   flags
 
@@ -86,44 +119,90 @@ decor_version (void)
   9rd and 10th bit alignment, 11rd and 12th bit clamp,
   13th bit XX, 14th bit XY, 15th bit YX, 16th bit YY.
 
-  data[12 + n * 9 + 1] = flags
-  data[12 + n * 9 + 2] = p1 x
-  data[12 + n * 9 + 3] = p1 y
-  data[12 + n * 9 + 4] = p2 x
-  data[12 + n * 9 + 5] = p2 y
-  data[12 + n * 9 + 6] = widthMax
-  data[12 + n * 9 + 7] = heightMax
-  data[12 + n * 9 + 8] = x0
-  data[12 + n * 9 + 9] = y0
+  data[26 + n * 9 + 1] = flags
+  data[26 + n * 9 + 2] = p1 x
+  data[26 + n * 9 + 3] = p1 y
+  data[26 + n * 9 + 4] = p2 x
+  data[26 + n * 9 + 5] = p2 y
+  data[26 + n * 9 + 6] = widthMax
+  data[26 + n * 9 + 7] = heightMax
+  data[26 + n * 9 + 8] = x0
+  data[26 + n * 9 + 9] = y0
  */
+
+long *
+decor_alloc_property (unsigned int n,
+                      unsigned int type)
+{
+    unsigned int  propSize = 0;
+    long          *data;
+
+    if (type == WINDOW_DECORATION_TYPE_WINDOW)
+        propSize = WINDOW_PROP_SIZE;
+    else if (type == WINDOW_DECORATION_TYPE_PIXMAP)
+        propSize = BASE_PROP_SIZE + N_QUADS_MAX * QUAD_PROP_SIZE;
+
+    propSize *= n;
+    propSize += PROP_HEADER_SIZE;
+
+    data = calloc (propSize, sizeof (long));
+
+    data[0]  = DECOR_INTERFACE_VERSION;
+    data[1]  = type;
+    data[2]  = n;
+
+    return data;
+}
+
 void
 decor_quads_to_property (long		 *data,
+			 unsigned int    n,
 			 Pixmap		 pixmap,
-			 decor_extents_t *input,
-			 decor_extents_t *max_input,
+			 decor_extents_t *frame,
+			 decor_extents_t *border,
+			 decor_extents_t *max_frame,
+			 decor_extents_t *max_border,
 			 int		 min_width,
 			 int		 min_height,
 			 decor_quad_t    *quad,
-			 int		 nQuad)
+			 int		 nQuad,
+			 unsigned int	 frame_type,
+			 unsigned int    frame_state,
+			 unsigned int    frame_actions)
 {
-    *data++ = DECOR_INTERFACE_VERSION;
-    *data++ = WINDOW_DECORATION_TYPE_PIXMAP;
+    /* FIXME: Allocating for N_QUAD_MAX is slightly inefficient, but there
+     * isn't really a better way to do this at the moment */
+    data += PROP_HEADER_SIZE + n * (BASE_PROP_SIZE + QUAD_PROP_SIZE * N_QUADS_MAX);
 
     memcpy (data++, &pixmap, sizeof (Pixmap));
 
-    *data++ = input->left;
-    *data++ = input->right;
-    *data++ = input->top;
-    *data++ = input->bottom;
+    *data++ = frame->left;
+    *data++ = frame->right;
+    *data++ = frame->top;
+    *data++ = frame->bottom;
+    *data++ = border->left;
+    *data++ = border->right;
+    *data++ = border->top;
+    *data++ = border->bottom;
 
-    *data++ = max_input->left;
-    *data++ = max_input->right;
-    *data++ = max_input->top;
-    *data++ = max_input->bottom;
+    *data++ = max_frame->left;
+    *data++ = max_frame->right;
+    *data++ = max_frame->top;
+    *data++ = max_frame->bottom;
+    *data++ = max_border->left;
+    *data++ = max_border->right;
+    *data++ = max_border->top;
+    *data++ = max_border->bottom;
 
     *data++ = min_width;
     *data++ = min_height;
 
+    *data++ = frame_type;
+    *data++ = frame_state;
+    *data++ = frame_actions;
+
+    *data++ = nQuad;
+
     while (nQuad--)
     {
 	*data++ =
@@ -152,13 +231,16 @@ decor_quads_to_property (long		 *data,
 
 void
 decor_gen_window_property (long		   *data,
+			   unsigned int    n,
 			   decor_extents_t *input,
 			   decor_extents_t *max_input,
 			   int		   min_width,
-			   int		   min_height)
+			   int		   min_height,
+			   unsigned int	   frame_type,
+			   unsigned int    frame_state,
+			   unsigned int    frame_actions)
 {
-    *data++ = DECOR_INTERFACE_VERSION;
-    *data++ = WINDOW_DECORATION_TYPE_WINDOW;
+    data += PROP_HEADER_SIZE + n * WINDOW_PROP_SIZE;
 
     *data++ = input->left;
     *data++ = input->right;
@@ -172,6 +254,10 @@ decor_gen_window_property (long		   *data,
 
     *data++ = min_width;
     *data++ = min_height;
+
+    *data++ = frame_type;
+    *data++ = frame_state;
+    *data++ = frame_actions;
 }
 
 int
@@ -186,46 +272,69 @@ decor_property_get_type (long *data)
     return (int) data[1];
 }
 
+int
+decor_property_get_num (long *data)
+{
+    return (int) data[2];
+}
+
 int
 decor_pixmap_property_to_quads (long		 *data,
+				unsigned int     nOffset,
 				int		 size,
 				Pixmap		 *pixmap,
-				decor_extents_t *input,
-				decor_extents_t *max_input,
+				decor_extents_t  *frame,
+				decor_extents_t  *border,
+				decor_extents_t  *max_frame,
+				decor_extents_t  *max_border,
 				int		 *min_width,
 				int		 *min_height,
-				decor_quad_t    *quad)
+				unsigned int     *frame_type,
+				unsigned int     *frame_state,
+				unsigned int     *frame_actions,
+				decor_quad_t     *quad)
 {
     int i, n, flags;
 
-    if (size < BASE_PROP_SIZE + QUAD_PROP_SIZE)
-	return 0;
+    if (size < PROP_HEADER_SIZE + nOffset * (BASE_PROP_SIZE + QUAD_PROP_SIZE + N_QUADS_MAX))
+        return 0;
 
     if (decor_property_get_version (data) != decor_version ())
-	return 0;
+        return 0;
 
     if (decor_property_get_type (data) != WINDOW_DECORATION_TYPE_PIXMAP)
-	return 0;
+        return 0;
 
-    data++;
-    data++;
+    data += PROP_HEADER_SIZE + nOffset * (BASE_PROP_SIZE + N_QUADS_MAX * QUAD_PROP_SIZE);
 
     memcpy (pixmap, data++, sizeof (Pixmap));
 
-    input->left   = *data++;
-    input->right  = *data++;
-    input->top    = *data++;
-    input->bottom = *data++;
+    frame->left   = *data++;
+    frame->right  = *data++;
+    frame->top    = *data++;
+    frame->bottom = *data++;
+    border->left   = *data++;
+    border->right  = *data++;
+    border->top    = *data++;
+    border->bottom = *data++;
 
-    max_input->left   = *data++;
-    max_input->right  = *data++;
-    max_input->top    = *data++;
-    max_input->bottom = *data++;
+    max_frame->left   = *data++;
+    max_frame->right  = *data++;
+    max_frame->top    = *data++;
+    max_frame->bottom = *data++;
+    max_border->left   = *data++;
+    max_border->right  = *data++;
+    max_border->top    = *data++;
+    max_border->bottom = *data++;
 
     *min_width  = *data++;
     *min_height = *data++;
 
-    n = (size - BASE_PROP_SIZE) / QUAD_PROP_SIZE;
+    *frame_type = *data++;
+    *frame_state = *data++;
+    *frame_actions = *data++;
+
+    n = *data++;
 
     for (i = 0; i < n; i++)
     {
@@ -260,25 +369,136 @@ decor_pixmap_property_to_quads (long		 *data,
     return n;
 }
 
+static int
+decor_point_cmp (const decor_point_t *a, const decor_point_t *b)
+{
+    /* Use binary | to avoid branch prediction slow-downs */
+    return (a->x - b->x) | (a->y - b->y) | (a->gravity - b->gravity);
+}
+
+static int
+decor_matrix_cmp (const decor_matrix_t *a, const decor_matrix_t *b)
+{
+    return (a->xx != b->xx) ||
+           (a->yx != b->yx) ||
+           (a->xy != b->xy) ||
+           (a->yy != b->yy) ||
+           (a->x0 != b->x0) ||
+           (a->y0 != b->y0);
+}
+
+static int
+decor_quad_cmp (const decor_quad_t *a, const decor_quad_t *b)
+{
+    return decor_point_cmp (&a->p1, &b->p1) ||
+           decor_point_cmp (&a->p2, &b->p2) ||
+           decor_matrix_cmp (&a->m, &b->m)  ||
+           (
+               (a->max_width  - b->max_width)  |
+               (a->max_height - b->max_height) |
+               (a->align      - b->align)      |
+               (a->clamp      - b->clamp)      |
+               (a->stretch    - b->stretch)
+           );
+}
+
+static int
+decor_extents_cmp (const decor_extents_t *a, const decor_extents_t *b)
+{
+    /* Use binary | to avoid branch prediction slow-downs */
+    return (a->left   - b->left)  |
+           (a->right  - b->right) |
+           (a->top    - b->top)   |
+           (a->bottom - b->bottom);
+}
+
+/* Returns n for a match, returns -1 for no match */
+
+int
+decor_match_pixmap (long		 *data,
+		    int		 size,
+		    Pixmap		 *pixmap,
+		    decor_extents_t  *frame,
+		    decor_extents_t  *border,
+		    decor_extents_t  *max_frame,
+		    decor_extents_t  *max_border,
+		    int		 min_width,
+		    int		 min_height,
+		    unsigned int     frame_type,
+		    unsigned int     frame_state,
+		    unsigned int     frame_actions,
+		    decor_quad_t     *quad,
+		    unsigned int     n_quad)
+{
+    int n = decor_property_get_num (data);
+    unsigned int i = 0;
+
+    for (; i < n; i++)
+    {
+	Pixmap cPixmap;
+	decor_extents_t cFrame, cBorder, cMax_frame, cMax_border;
+	int cMin_width, cMin_height;
+	int q;
+	unsigned int cFrame_type, cFrame_state, cFrame_actions, cNQuad;
+	decor_quad_t cQuad[N_QUADS_MAX];
+	cNQuad = decor_pixmap_property_to_quads (data, i, size, &cPixmap, &cFrame, &cBorder, &cMax_frame,
+						 &cMax_border, &cMin_width, &cMin_height, &cFrame_type,
+						 &cFrame_state, &cFrame_actions, cQuad);
+
+	if (cPixmap != *pixmap)
+	    continue;
+
+	if (decor_extents_cmp (&cFrame, frame) ||
+	    decor_extents_cmp (&cBorder, border) ||
+	    decor_extents_cmp (&cMax_frame, max_frame) ||
+	    decor_extents_cmp (&cMax_border, max_border))
+	    continue;
+
+	if (cFrame_type != frame_type ||
+	    cFrame_state != frame_state ||
+	    cFrame_actions != frame_actions ||
+	    cMin_width != min_width ||
+	    cMin_height != min_height)
+	    continue;
+
+	if (cNQuad != n_quad)
+	    continue;
+
+	q = 0;
+	while (q < n_quad && !decor_quad_cmp (&cQuad[q], &quad[q]))
+	    q++;
+
+	if (q < n_quad)
+	    continue;
+
+	return n;
+    }
+
+    return -1;
+}
+
 int
 decor_window_property (long	       *data,
+		       unsigned int    n,
 		       int	       size,
 		       decor_extents_t *input,
 		       decor_extents_t *max_input,
 		       int	       *min_width,
-		       int	       *min_height)
+		       int	       *min_height,
+		       unsigned int    *frame_type,
+		       unsigned int    *frame_state,
+		       unsigned int    *frame_actions)
 {
-    if (size < WINDOW_PROP_SIZE)
-	return 0;
-
     if (decor_property_get_version (data) != decor_version ())
-	return 0;
+        return 0;
 
     if (decor_property_get_type (data) != WINDOW_DECORATION_TYPE_WINDOW)
-	return 0;
+        return 0;
 
-    data++;
-    data++;
+    if (size < PROP_HEADER_SIZE + n * WINDOW_PROP_SIZE)
+        return 0;
+
+    data += PROP_HEADER_SIZE + n * WINDOW_PROP_SIZE;
 
     input->left   = *data++;
     input->right  = *data++;
@@ -293,6 +513,10 @@ decor_window_property (long	       *data,
     *min_width  = *data++;
     *min_height = *data++;
 
+    *frame_type = *data++;
+    *frame_state = *data++;
+    *frame_actions = *data++;
+
     return 1;
 }
 
@@ -1642,6 +1866,8 @@ decor_get_best_layout (decor_context_t *c,
 	}
 	else
 	{
+	    layout->rotation = 0;
+
 	    layout->top.pad    = 0;
 	    layout->bottom.pad = 0;
 	    layout->left.pad   = 0;
