@@ -23,11 +23,11 @@
  * Author: David Reveman <davidr@novell.com>
  */
 
-#include <core/core.h>
-#include <core/propertywriter.h>
 #include <compiztoolbox/compiztoolbox.h>
 #include "compiztoolbox_options.h"
 
+#include <core/abiversion.h>
+#include <core/propertywriter.h>
 
 bool openGLAvailable;
 
@@ -42,11 +42,12 @@ class CompizToolboxScreen :
 class CompizToolboxPluginVTable :
     public CompPlugin::VTableForScreen <CompizToolboxScreen>
 {
-    bool init ();
-    void fini ();
+    public:
+	bool init ();
+	void fini ();
 };
 
-COMPIZ_PLUGIN_20090315 (compiztoolbox, CompizToolboxPluginVTable);
+COMPIZ_PLUGIN_20090315 (compiztoolbox, CompizToolboxPluginVTable)
 
 CompString
 getXDGUserDir (XDGUserDir userDir)
@@ -122,15 +123,24 @@ getXDGUserDir (XDGUserDir userDir)
 
 
 void
-BaseSwitchScreen::setSelectedWindowHint ()
+BaseSwitchScreen::setSelectedWindowHint (bool focus)
 {
     Window selectedWindowId = None;
     CompOption::Vector opts;
     CompOption::Value  v;
 
     if (selectedWindow && !selectedWindow->destroyed ())
+    {
 	selectedWindowId = selectedWindow->id ();
 
+	/* FIXME: Changing the input focus here will
+	 * screw up the ordering of windows in
+	 * the switcher, so we probably want to avoid that
+	 */
+	if (focus)
+	    selectedWindow->moveInputFocusTo ();
+    }
+
     v = CompOption::Value ((int) selectedWindowId);
     opts = selectWinAtom.getReadTemplate ();
     opts.at (0).set (v);
@@ -245,7 +255,8 @@ BaseSwitchScreen::compareWindows (CompWindow *w1,
 
 CompWindow *
 BaseSwitchScreen::switchToWindow (bool toNext,
-				  bool autoChangeVPOption)
+				  bool autoChangeVPOption,
+				  bool focus)
 {
     CompWindow               *w = NULL;
     CompWindowList::iterator it;
@@ -324,7 +335,7 @@ BaseSwitchScreen::switchToWindow (bool toNext,
 	    if (popup)
 		CompositeWindow::get (popup)->addDamage ();
 
-	    setSelectedWindowHint ();
+	    setSelectedWindowHint (focus);
 	}
 
 	doWindowDamage (w);
@@ -393,7 +404,7 @@ BaseSwitchWindow::paintThumb (const GLWindowPaintAttrib &attrib,
 			      int                       height2)
 {
     if (!openGLAvailable)
-	return;    
+	return;
 
     GLWindowPaintAttrib  sAttrib (attrib);
     IconMode             iconMode;
@@ -426,8 +437,8 @@ BaseSwitchWindow::paintThumb (const GLWindowPaintAttrib &attrib,
 	width  = width1;
 	height = height1;
 
-	ww = window->inputRect ().width ();
-	wh = window->inputRect ().height ();
+	ww = window->borderRect ().width ();
+	wh = window->borderRect ().height ();
 
 	if (ww > width)
 	    sAttrib.xScale = width / ww;
@@ -450,9 +461,9 @@ BaseSwitchWindow::paintThumb (const GLWindowPaintAttrib &attrib,
 	updateIconPos (wx, wy, x, y, width, height);
 
 	sAttrib.xTranslate = wx - g.x () +
-			     window->input ().left * sAttrib.xScale;
+			     window->border ().left * sAttrib.xScale;
 	sAttrib.yTranslate = wy - g.y () +
-			     window->input ().top  * sAttrib.yScale;
+			     window->border ().top  * sAttrib.yScale;
 
 	GLFragment::Attrib fragment (sAttrib);
 
@@ -683,7 +694,7 @@ BaseSwitchScreen::BaseSwitchScreen (CompScreen *screen) :
     selectWinAtom = PropertyWriter (CompString (DECOR_SWITCH_WINDOW_ATOM_NAME),
     				    atomTemplate);
 
-    selectFgColorAtom = 
+    selectFgColorAtom =
     	XInternAtom (::screen->dpy (),
     		     DECOR_SWITCH_FOREGROUND_COLOR_ATOM_NAME, 0);
 
