@@ -43,8 +43,9 @@
 #include <X11/extensions/shape.h>
 #include <X11/extensions/Xrandr.h>
 
-CompWindow *lastDamagedWindow = 0;
+#include <core/timer.h>
 
+CompWindow *lastDamagedWindow = 0;
 
 void
 PrivateCompositeScreen::handleEvent (XEvent *event)
@@ -259,13 +260,8 @@ CompositeScreen::CompositeScreen (CompScreen *s) :
 CompositeScreen::~CompositeScreen ()
 {
     priv->paintTimer.stop ();
-
-#ifdef USE_COW
-    if (useCow)
-	XCompositeReleaseOverlayWindow (screen->dpy (),
-					screen->root ());
-#endif
-
+    XCompositeReleaseOverlayWindow (screen->dpy (),
+				    screen->root ());
     delete priv;
 }
 
@@ -278,18 +274,17 @@ PrivateCompositeScreen::PrivateCompositeScreen (CompositeScreen *cs) :
     exposeRects (),
     windowPaintOffset (0, 0),
     overlayWindowCount (0),
-    nextRedraw (0),
     redrawTime (1000 / 50),
     optimalRedrawTime (1000 / 50),
-    frameStatus (0),
-    timeMult (1),
-    idle (true),
-    timeLeft (0),
+    scheduled (false),
+    painting (false),
+    reschedule (false),
     slowAnimations (false),
-    active (false),
     pHnd (NULL),
     FPSLimiterMode (CompositeFPSLimiterModeDefault),
-    frameTimeAccumulator (0)
+    withDestroyedWindows (),
+    cmSnAtom (0),
+    newCmSnOwner (None)
 {
     gettimeofday (&lastRedraw, 0);
     // wrap outputChangeNotify
@@ -300,14 +295,19 @@ PrivateCompositeScreen::PrivateCompositeScreen (CompositeScreen *cs) :
 
 PrivateCompositeScreen::~PrivateCompositeScreen ()
 {
+    Display *dpy = screen->dpy ();
+
+    if (cmSnAtom)
+	XSetSelectionOwner (dpy, cmSnAtom, None, CurrentTime);
+
+    if (newCmSnOwner != None)
+	XDestroyWindow (dpy, newCmSnOwner);
 }
 
 bool
 PrivateCompositeScreen::init ()
 {
     Display              *dpy = screen->dpy ();
-    Window               newCmSnOwner = None;
-    Atom                 cmSnAtom = 0;
     Time                 cmSnTimestamp = 0;
     XEvent               event;
     XSetWindowAttributes attr;
@@ -323,12 +323,12 @@ PrivateCompositeScreen::init ()
     {
 	if (!replaceCurrentWm)
 	{
-	    compLogMessage ("composite", CompLogLevelError,
-			    "Screen %d on display \"%s\" already "
-			    "has a compositing manager; try using the "
-			    "--replace option to replace the current "
-			    "compositing manager.",
-			    screen->screenNum (), DisplayString (dpy));
+	    compLogMessage (
+		"composite", CompLogLevelError,
+		"Screen %d on display \"%s\" already has a compositing "
+		"manager (%x); try using the --replace option to replace "
+		"the current compositing manager.",
+		screen->screenNum (), DisplayString (dpy), currentCmSnOwner);
 
 	    return false;
 	}
@@ -383,11 +383,15 @@ PrivateCompositeScreen::init ()
 
 
 bool
-CompositeScreen::registerPaintHandler (PaintHandler *pHnd)
+CompositeScreen::registerPaintHandler (compiz::composite::PaintHandler *pHnd)
 {
-    Display *dpy = screen->dpy ();
+    Display *dpy;
 
-    if (priv->active)
+    WRAPABLE_HND_FUNCTN_RETURN (bool, registerPaintHandler, pHnd);
+
+    dpy =  screen->dpy ();
+
+    if (priv->pHnd)
 	return false;
 
     CompScreen::checkForError (dpy);
@@ -414,21 +418,20 @@ CompositeScreen::registerPaintHandler (PaintHandler *pHnd)
     }
 
     priv->pHnd = pHnd;
-    priv->active = true;
 
     showOutputWindow ();
 
-    priv->paintTimer.start
-	(boost::bind (&CompositeScreen::handlePaintTimeout, this),
-	 priv->optimalRedrawTime, MAXSHORT);
-
     return true;
 }
 
 void
 CompositeScreen::unregisterPaintHandler ()
 {
-    Display *dpy = screen->dpy ();
+    Display *dpy;
+
+    WRAPABLE_HND_FUNCTN (unregisterPaintHandler)
+
+    dpy = screen->dpy ();
 
     foreach (CompWindow *w, screen->windows ())
     {
@@ -444,7 +447,6 @@ CompositeScreen::unregisterPaintHandler ()
 				    CompositeRedirectManual);
 
     priv->pHnd = NULL;
-    priv->active = false;
     priv->paintTimer.stop ();
 
     hideOutputWindow ();
@@ -453,30 +455,40 @@ CompositeScreen::unregisterPaintHandler ()
 bool
 CompositeScreen::compositingActive ()
 {
-    return priv->active;
+    if (priv->pHnd)
+	return priv->pHnd->compositingActive ();
+
+    return false;
 }
 
 void
 CompositeScreen::damageScreen ()
 {
-    if (priv->damageMask == 0)
-	priv->paintTimer.setTimes (priv->paintTimer.minLeft ());
+    bool alreadyDamaged = priv->damageMask & COMPOSITE_SCREEN_DAMAGE_ALL_MASK;
 
     priv->damageMask |= COMPOSITE_SCREEN_DAMAGE_ALL_MASK;
     priv->damageMask &= ~COMPOSITE_SCREEN_DAMAGE_REGION_MASK;
+
+    priv->scheduleRepaint ();
+
+    /*
+     * Call through damageRegion since plugins listening for incoming damage
+     * may need to know that the whole screen was redrawn
+     */
+
+    if (!alreadyDamaged)
+	damageRegion (CompRegion (0, 0, screen->width (), screen->height ()));
 }
 
 void
 CompositeScreen::damageRegion (const CompRegion &region)
 {
+    WRAPABLE_HND_FUNCTN (damageRegion, region);
+
     if (priv->damageMask & COMPOSITE_SCREEN_DAMAGE_ALL_MASK)
 	return;
 
-    if (priv->damageMask == 0)
-	priv->paintTimer.setTimes (priv->paintTimer.minLeft ());
-
     priv->damage += region;
-
     priv->damageMask |= COMPOSITE_SCREEN_DAMAGE_REGION_MASK;
 
     /* if the number of damage rectangles grows two much between repaints,
@@ -486,15 +498,14 @@ CompositeScreen::damageRegion (const CompRegion &region)
 
     if (priv->damage.numRects () > 100)
        damageScreen ();
+    priv->scheduleRepaint ();
 }
 
 void
 CompositeScreen::damagePending ()
 {
-    if (priv->damageMask == 0)
-	priv->paintTimer.setTimes (priv->paintTimer.minLeft ());
-
     priv->damageMask |= COMPOSITE_SCREEN_DAMAGE_PENDING_MASK;
+    priv->scheduleRepaint ();
 }
 
 unsigned int
@@ -506,8 +517,7 @@ CompositeScreen::damageMask ()
 void
 CompositeScreen::showOutputWindow ()
 {
-#ifdef USE_COW
-    if (useCow && priv->active)
+    if (priv->pHnd)
     {
 	Display       *dpy = screen->dpy ();
 	XserverRegion region;
@@ -527,37 +537,28 @@ CompositeScreen::showOutputWindow ()
 
 	damageScreen ();
     }
-#endif
-
 }
 
 void
 CompositeScreen::hideOutputWindow ()
 {
-#ifdef USE_COW
-    if (useCow)
-    {
-	Display       *dpy = screen->dpy ();
-	XserverRegion region;
+    Display       *dpy = screen->dpy ();
+    XserverRegion region;
 
-	region = XFixesCreateRegion (dpy, NULL, 0);
+    region = XFixesCreateRegion (dpy, NULL, 0);
 
-	XFixesSetWindowShapeRegion (dpy,
-				    priv->output,
-				    ShapeBounding,
-				    0, 0, region);
-
-	XFixesDestroyRegion (dpy, region);
-    }
-#endif
+    XFixesSetWindowShapeRegion (dpy,
+				priv->output,
+				ShapeBounding,
+				0, 0, region);
 
+    XFixesDestroyRegion (dpy, region);
 }
 
 void
 CompositeScreen::updateOutputWindow ()
 {
-#ifdef USE_COW
-    if (useCow && priv->active)
+    if (priv->pHnd)
     {
 	Display       *dpy = screen->dpy ();
 	XserverRegion region;
@@ -584,24 +585,16 @@ CompositeScreen::updateOutputWindow ()
 
 	XFixesDestroyRegion (dpy, region);
     }
-#endif
 
 }
 
 void
 PrivateCompositeScreen::makeOutputWindow ()
 {
-#ifdef USE_COW
-    if (useCow)
-    {
-	overlay = XCompositeGetOverlayWindow (screen->dpy (), screen->root ());
-	output  = overlay;
+    overlay = XCompositeGetOverlayWindow (screen->dpy (), screen->root ());
+    output  = overlay;
 
-	XSelectInput (screen->dpy (), output, ExposureMask);
-    }
-    else
-#endif
-	output = overlay = screen->root ();
+    XSelectInput (screen->dpy (), output, ExposureMask);
 
     cScreen->hideOutputWindow ();
 }
@@ -639,15 +632,14 @@ CompositeScreen::windowPaintOffset ()
 void
 PrivateCompositeScreen::detectRefreshRate ()
 {
-    if (!noDetection &&
-	optionGetDetectRefreshRate ())
+    if (optionGetDetectRefreshRate ())
     {
 	CompString        name;
 	CompOption::Value value;
 
 	value.set ((int) 0);
 
-	if (screen->XRandr ())
+	if (randrExtension)
 	{
 	    XRRScreenConfiguration *config;
 
@@ -664,6 +656,7 @@ PrivateCompositeScreen::detectRefreshRate ()
 	mOptions[CompositeOptions::DetectRefreshRate].value ().set (false);
 	screen->setOptionForPlugin ("composite", "refresh_rate", value);
 	mOptions[CompositeOptions::DetectRefreshRate].value ().set (true);
+	optimalRedrawTime = redrawTime = 1000 / value.i ();
     }
     else
     {
@@ -684,75 +677,39 @@ CompositeScreen::setFPSLimiterMode (CompositeFPSLimiterMode newMode)
     priv->FPSLimiterMode = newMode;
 }
 
-int
-PrivateCompositeScreen::getTimeToNextRedraw (struct timeval *tv)
+void
+PrivateCompositeScreen::scheduleRepaint ()
 {
-    int diff, next;
-
-    diff = TIMEVALDIFF (tv, &lastRedraw);
-
-    /* handle clock rollback */
-    if (diff < 0)
-	diff = 0;
-
-    bool hasVSyncBehavior =
-	(FPSLimiterMode == CompositeFPSLimiterModeVSyncLike ||
-	 (pHnd && pHnd->hasVSync ()));
-
-    if (idle || hasVSyncBehavior)
+    if (painting)
     {
-	if (timeMult > 1)
-	{
-	    frameStatus = -1;
-	    redrawTime = optimalRedrawTime;
-	    timeMult--;
-	}
+	reschedule = true;
+	return;
+    }
+
+    if (scheduled)
+	return;
+
+    scheduled = true;
+
+    int delay;
+    if (FPSLimiterMode == CompositeFPSLimiterModeVSyncLike ||
+	(pHnd && pHnd->hasVSync ()))
+    {
+	delay = 1;
     }
     else
     {
-	if (diff > redrawTime)
-	{
-	    if (frameStatus > 0)
-		frameStatus = 0;
-
-	    next = optimalRedrawTime * (timeMult + 1);
-	    if (diff > next)
-	    {
-		frameStatus--;
-		if (frameStatus < -1)
-		{
-		    timeMult++;
-		    redrawTime = diff = next;
-		}
-	    }
-	}
-	else if (diff < redrawTime)
-	{
-	    if (frameStatus < 0)
-		frameStatus = 0;
-
-	    if (timeMult > 1)
-	    {
-		next = optimalRedrawTime * (timeMult - 1);
-		if (diff < next)
-		{
-		    frameStatus++;
-		    if (frameStatus > 4)
-		    {
-			timeMult--;
-			redrawTime = next;
-		    }
-		}
-	    }
-	}
+	struct timeval now;
+	gettimeofday (&now, 0);
+	int elapsed = compiz::core::timer::timeval_diff (&now, &lastRedraw);
+	if (elapsed < 0)
+	    elapsed = 0;
+ 	delay = elapsed < optimalRedrawTime ? optimalRedrawTime - elapsed : 1;
     }
-    if (diff >= redrawTime)
-	return 1;
 
-    if (hasVSyncBehavior)
-	return (redrawTime - diff) * 0.7;
-
-    return redrawTime - diff;
+    paintTimer.start
+	(boost::bind (&CompositeScreen::handlePaintTimeout, cScreen),
+	delay);
 }
 
 int
@@ -770,37 +727,36 @@ CompositeScreen::optimalRedrawTime ()
 bool
 CompositeScreen::handlePaintTimeout ()
 {
-    int         timeDiff;
     struct      timeval tv;
-    int         timeToNextRedraw;
 
+    priv->painting = true;
+    priv->reschedule = false;
     gettimeofday (&tv, 0);
 
     if (priv->damageMask)
     {
+	int         timeDiff;
+
 	if (priv->pHnd)
 	    priv->pHnd->prepareDrawing ();
 
-	timeDiff = TIMEVALDIFF (&tv, &priv->lastRedraw);
+	timeDiff = compiz::core::timer::timeval_diff (&tv, &priv->lastRedraw);
 
 	/* handle clock rollback */
 	if (timeDiff < 0)
 	    timeDiff = 0;
+	/*
+	 * Now that we use a "tickless" timing algorithm, timeDiff could be
+	 * very large if the screen is truely idle.
+	 * However plugins expect the old behaviour where timeDiff is rarely
+	 * larger than the frame rate (optimalRedrawTime).
+	 * So enforce this to keep animations timed correctly and smooth...
+	 */
+	if (timeDiff > 100)
+	    timeDiff = priv->optimalRedrawTime;
 
-	if (priv->slowAnimations)
-	{
-	    int msSinceLastPaint;
-
-	    if (priv->FPSLimiterMode == CompositeFPSLimiterModeDisabled)
-		msSinceLastPaint = 1;
-	    else
-		msSinceLastPaint =
-		    priv->idle ? 2 : (timeDiff * 2) / priv->redrawTime;
-
-	    preparePaint (msSinceLastPaint);
-	}
-	else
-	    preparePaint (priv->idle ? priv->redrawTime : timeDiff);
+	priv->redrawTime = timeDiff;
+	preparePaint (priv->slowAnimations ? 1 : timeDiff);
 
 	/* substract top most overlay window region */
 	if (priv->overlayWindowCount)
@@ -853,7 +809,6 @@ CompositeScreen::handlePaintTimeout ()
 
 	paint (outputs, mask);
 
-	priv->lastRedraw = tv;
 
 	donePaint ();
 
@@ -865,55 +820,30 @@ CompositeScreen::handlePaintTimeout ()
 		break;
 	    }
 	}
-
-	priv->idle = false;
-    }
-    else
-    {
-	priv->idle = true;
     }
 
-    gettimeofday (&tv, 0);
+    priv->lastRedraw = tv;
+    priv->painting = false;
+    priv->scheduled = false;
+    if (priv->reschedule)
+	priv->scheduleRepaint ();
 
-    if (priv->FPSLimiterMode == CompositeFPSLimiterModeDisabled)
-    {
-	if (priv->FPSLimiterMode == CompositeFPSLimiterModeDisabled)
-	{
-	    const int msToReturn1After = 100;
-
-	    priv->frameTimeAccumulator += priv->redrawTime;
-	    if (priv->frameTimeAccumulator > msToReturn1After)
-	    {
-		priv->frameTimeAccumulator %= msToReturn1After;
-		timeToNextRedraw = 1;
-	    }
-	    else
-		timeToNextRedraw = 0;
-	}
-    }
-    else
-	timeToNextRedraw = priv->getTimeToNextRedraw (&tv);
-
-    if (priv->idle)
-	priv->paintTimer.setTimes (timeToNextRedraw, MAXSHORT);
-    else
-	priv->paintTimer.setTimes (timeToNextRedraw);
-    return true;
+    return false;
 }
 
 void
 CompositeScreen::preparePaint (int msSinceLastPaint)
-    WRAPABLE_HND_FUNC (0, preparePaint, msSinceLastPaint)
+    WRAPABLE_HND_FUNCTN (preparePaint, msSinceLastPaint)
 
 void
 CompositeScreen::donePaint ()
-    WRAPABLE_HND_FUNC (1, donePaint)
+    WRAPABLE_HND_FUNCTN (donePaint)
 
 void
 CompositeScreen::paint (CompOutput::ptrList &outputs,
 		        unsigned int        mask)
 {
-    WRAPABLE_HND_FUNC (2, paint, outputs, mask)
+    WRAPABLE_HND_FUNCTN (paint, outputs, mask)
 
     if (priv->pHnd)
 	priv->pHnd->paintOutputs (outputs, mask, priv->tmpRegion);
@@ -922,9 +852,41 @@ CompositeScreen::paint (CompOutput::ptrList &outputs,
 const CompWindowList &
 CompositeScreen::getWindowPaintList ()
 {
-    WRAPABLE_HND_FUNC_RETURN (3, const CompWindowList &, getWindowPaintList)
+    WRAPABLE_HND_FUNCTN_RETURN (const CompWindowList &, getWindowPaintList)
 
-    return screen->windows ();
+    /* Include destroyed windows */
+    if (screen->destroyedWindows ().empty ())
+	return screen->windows ();
+    else
+    {
+	CompWindowList destroyedWindows = screen->destroyedWindows ();
+
+	priv->withDestroyedWindows.resize (0);
+
+	foreach (CompWindow *w, screen->windows ())
+	{
+	    foreach (CompWindow *dw, screen->destroyedWindows ())
+	    {
+		if (dw->next == w)
+		{
+		    priv->withDestroyedWindows.push_back (dw);
+		    destroyedWindows.remove (dw);
+		    break;
+		}
+	    }
+
+	    priv->withDestroyedWindows.push_back (w);
+	}
+
+	/* We need to put all the destroyed windows which didn't get
+	 * inserted in the paint list at the top of the stack since
+	 * w->next was probably either invalid or NULL */
+
+	foreach (CompWindow *dw, destroyedWindows)
+	    priv->withDestroyedWindows.push_back (dw);
+
+	return priv->withDestroyedWindows;
+    }
 }
 
 void
@@ -953,11 +915,8 @@ void
 PrivateCompositeScreen::outputChangeNotify ()
 {
     screen->outputChangeNotify ();
-#ifdef USE_COW
-    if (useCow)
-	XMoveResizeWindow (screen->dpy (), overlay, 0, 0,
-			   screen->width (), screen->height ());
-#endif
+    XMoveResizeWindow (screen->dpy (), overlay, 0, 0,
+		       screen->width (), screen->height ());
     cScreen->damageScreen ();
 }
 
@@ -991,6 +950,18 @@ const CompWindowList &
 CompositeScreenInterface::getWindowPaintList ()
     WRAPABLE_DEF (getWindowPaintList)
 
+bool
+CompositeScreenInterface::registerPaintHandler (compiz::composite::PaintHandler *pHnd)
+    WRAPABLE_DEF (registerPaintHandler, pHnd);
+
+void
+CompositeScreenInterface::unregisterPaintHandler ()
+    WRAPABLE_DEF (unregisterPaintHandler);
+
+void
+CompositeScreenInterface::damageRegion (const CompRegion &r)
+    WRAPABLE_DEF (damageRegion, r);
+
 const CompRegion &
 CompositeScreen::currentDamage () const
 {
