@@ -31,31 +31,32 @@
 #include <X11/Xatom.h>
 #include <X11/cursorfont.h>
 
-#include <core/core.h>
 #include <core/atoms.h>
 #include "resize.h"
 
 COMPIZ_PLUGIN_20090315 (resize, ResizePluginVTable)
 
+#define XWINDOWCHANGES_INIT {0, 0, 0, 0, 0, None, 0}
+
 void
 ResizeScreen::getPaintRectangle (BoxPtr pBox)
 {
-    pBox->x1 = geometry.x - w->input ().left;
-    pBox->y1 = geometry.y - w->input ().top;
+    pBox->x1 = geometry.x - w->border ().left;
+    pBox->y1 = geometry.y - w->border ().top;
     pBox->x2 = geometry.x + geometry.width +
-	       w->serverGeometry ().border () * 2 + w->input ().right;
+	       w->serverGeometry ().border () * 2 + w->border ().right;
 
     if (w->shaded ())
-	pBox->y2 = geometry.y + w->size ().height () + w->input ().bottom;
+	pBox->y2 = geometry.y + w->size ().height () + w->border ().bottom;
     else
 	pBox->y2 = geometry.y + geometry.height +
-	           w->serverGeometry ().border () * 2 + w->input ().bottom;
+		   w->serverGeometry ().border () * 2 + w->border ().bottom;
 }
 
 void
 ResizeWindow::getStretchScale (BoxPtr pBox, float *xScale, float *yScale)
 {
-    CompRect rect (window->inputRect ());
+    CompRect rect (window->borderRect ());
 
     *xScale = (rect.width ())  ? (pBox->x2 - pBox->x1) /
 				 (float) rect.width () : 1.0f;
@@ -72,8 +73,8 @@ ResizeScreen::getStretchRectangle (BoxPtr pBox)
     getPaintRectangle (&box);
     ResizeWindow::get (w)->getStretchScale (&box, &xScale, &yScale);
 
-    pBox->x1 = (int) (box.x1 - (w->output ().left - w->input ().left) * xScale);
-    pBox->y1 = (int) (box.y1 - (w->output ().top - w->input ().top) * yScale);
+    pBox->x1 = (int) (box.x1 - (w->output ().left - w->border ().left) * xScale);
+    pBox->y1 = (int) (box.y1 - (w->output ().top - w->border ().top) * yScale);
     pBox->x2 = (int) (box.x2 + w->output ().right * xScale);
     pBox->y2 = (int) (box.y2 + w->output ().bottom * yScale);
 }
@@ -177,12 +178,128 @@ void
 ResizeScreen::finishResizing ()
 {
     w->ungrabNotify ();
-    
+
     resizeInformationAtom.deleteProperty (w->id ());
 
     w = NULL;
 }
 
+#define TOUCH_LEFT 1
+#define TOUCH_RIGHT 2
+#define TOUCH_TOP 3
+#define TOUCH_BOTTOM 4
+
+static unsigned int
+findTouchingOutput (int touchPoint, unsigned int side)
+{
+    for (unsigned int i = 0; i < screen->outputDevs ().size (); i++)
+    {
+	CompOutput &o = screen->outputDevs ().at (i);
+	if (side == TOUCH_LEFT)
+	{
+	    if (o.left () == touchPoint)
+		return  i;
+	}
+	if (side == TOUCH_RIGHT)
+	{
+	    if (o.right () == touchPoint)
+		return  i;
+	}
+	if (side == TOUCH_TOP)
+	{
+	    if (o.top () == touchPoint)
+		return  i;
+	}
+	if (side == TOUCH_BOTTOM)
+	{
+	    if (o.bottom () == touchPoint)
+		return  i;
+	}
+    }
+
+    return -1;
+}
+
+static void
+getPointForTp (unsigned int tp, unsigned int output, int &op, int &wap)
+{
+    CompRect og = CompRect (screen->outputDevs ().at (output));
+    CompRect wag = screen->outputDevs ().at (output).workArea ();
+
+    switch (tp)
+    {
+	case TOUCH_LEFT:
+	    op = og.right ();
+	    wap = wag.right ();
+	    break;
+	case TOUCH_RIGHT:
+	    op = og.left ();
+	    wap = wag.left ();
+	    break;
+	case TOUCH_TOP:
+	    op = og.bottom ();
+	    wap = wag.bottom ();
+	    break;
+	case TOUCH_BOTTOM:
+	    op = og.top ();
+	    wap = wag.top ();
+	    break;
+	default:
+	    return;
+    }
+}
+
+/* Be a little bit intelligent about how we calculate
+ * the workarea. Basically we want to be enclosed in
+ * any area that is obstructed by panels, but not
+ * where two outputs meet
+ *
+ * Also, it does not make sense to resize over
+ * non-touching outputs, so detect that case too
+ * */
+
+static int
+getOutputForEdge (int windowOutput, unsigned int touch, bool skipFirst)
+{
+    int op, wap;
+    int ret = windowOutput;
+
+    getPointForTp (touch, windowOutput, op, wap);
+
+    if ((op == wap) || skipFirst)
+    {
+	int co = windowOutput;
+
+	do
+	{
+	    int oco = co;
+
+	    co = findTouchingOutput (op, touch);
+
+	    /* Could not find a leftmost output from here
+	     * so we must have hit the edge of the universe */
+	    if (co == -1)
+	    {
+		ret = oco;
+		co = -1;
+		break;
+	    }
+
+	    getPointForTp (touch, co, op, wap);
+
+	    /* There is something in the way here.... */
+	    if (op != wap)
+	    {
+		ret = co;
+		co = -1;
+	    }
+	}
+        while (co != -1);
+    }
+
+    return ret;
+}
+
 static bool
 resizeInitiate (CompAction         *action,
 	        CompAction::State  state,
@@ -278,6 +395,8 @@ resizeInitiate (CompAction         *action,
 	rs->pointerDx = x - pointerX;
 	rs->pointerDy = y - pointerY;
 
+	rs->centered |= rs->optionGetResizeFromCenterMatch ().evaluate (w);
+
 	if ((w->state () & MAXIMIZE_STATE) == MAXIMIZE_STATE)
 	{
 	    /* if the window is fully maximized, showing the outline or
@@ -332,6 +451,13 @@ resizeInitiate (CompAction         *action,
 
 	    w->grabNotify (x, y, state, grabMask);
 
+	    /* Click raise happens implicitly on buttons 1, 2 and 3 so don't
+	     * restack this window again if the action buttonbinding was from
+	     * one of those buttons */
+	    if (screen->getOption ("raise_on_click")->value ().b () &&
+		button != Button1 && button != Button2 && button != Button3)
+		w->updateAttributes (CompStackingUpdateModeAboveFullscreen);
+
 	    /* using the paint rectangle is enough here
 	       as we don't have any stretch yet */
 	    rs->getPaintRectangle (&box);
@@ -349,11 +475,47 @@ resizeInitiate (CompAction         *action,
 
 	    rs->isConstrained = sourceExternalApp;
 
+	    /* Update offWorkAreaConstrained and workArea at grab time */
+	    rs->offWorkAreaConstrained = false;
 	    if (sourceExternalApp)
 	    {
+		int output = w->outputDevice ();
+		int lco, tco, bco, rco;
+		bool sl = screen->outputDevs ().at (output).workArea ().left () >
+			  w->serverGeometry ().left ();
+		bool sr = screen->outputDevs ().at (output).workArea ().right () <
+			  w->serverGeometry ().right ();
+		bool st = screen->outputDevs ().at (output).workArea ().top () >
+			  w->serverGeometry ().top ();
+		bool sb = screen->outputDevs ().at (output).workArea ().bottom () <
+			  w->serverGeometry ().bottom ();
+
+		lco = tco = bco = rco = output;
+
 		/* Prevent resizing beyond work area edges when resize is
 		   initiated externally (e.g. with window frame or menu)
 		   and not with a key (e.g. alt+button) */
+		rs->offWorkAreaConstrained = true;
+
+		lco = getOutputForEdge (output, TOUCH_RIGHT, sl);
+		rco = getOutputForEdge (output, TOUCH_LEFT, sr);
+		tco = getOutputForEdge (output, TOUCH_BOTTOM, st);
+		bco = getOutputForEdge (output, TOUCH_TOP, sb);
+
+		/* Now we need to form one big rect which describes
+		 * the available workarea */
+
+		int left = screen->outputDevs ().at (lco).workArea ().left ();
+		int right = screen->outputDevs ().at (rco).workArea ().right ();
+		int top = screen->outputDevs ().at (tco).workArea ().top ();
+		int bottom = screen->outputDevs ().at (bco).workArea ().bottom ();
+
+		rs->grabWindowWorkArea.reset (new CompRect (0, 0, 0, 0));
+		rs->grabWindowWorkArea->setLeft (left);
+		rs->grabWindowWorkArea->setRight (right);
+		rs->grabWindowWorkArea->setTop (top);
+		rs->grabWindowWorkArea->setBottom (bottom);
+
 
 		rs->inRegionStatus   = false;
 		rs->lastGoodHotSpotY = -1;
@@ -408,7 +570,7 @@ resizeTerminate (CompAction         *action,
     if (rs->w)
     {
 	CompWindow     *w = rs->w;
-	XWindowChanges xwc;
+	XWindowChanges xwc = XWINDOWCHANGES_INIT;
 	unsigned int   mask = 0;
 
 	if (rs->mode == ResizeOptions::ModeNormal)
@@ -508,7 +670,7 @@ ResizeScreen::updateWindowSize ()
     if (w->serverGeometry ().width ()  != geometry.width ||
 	w->serverGeometry ().height () != geometry.height)
     {
-	XWindowChanges xwc;
+	XWindowChanges xwc = XWINDOWCHANGES_INIT;
 
 	xwc.x	   = geometry.x;
 	xwc.y	   = geometry.y;
@@ -552,12 +714,12 @@ ResizeScreen::handleKeyEvent (KeyCode keycode)
 		int x, y, left, top, width, height;
 
 		CompWindow::Geometry server = w->serverGeometry ();
-		CompWindowExtents    input  = w->input ();
+		const CompWindowExtents    &border  = w->border ();
 
-		left   = server.x () - input.left;
-		top    = server.y () - input.top;
-		width  = input.left + server.width () + input.right;
-		height = input.top  + server.height () + input.bottom;
+		left   = server.x () - border.left;
+		top    = server.y () - border.top;
+		width  = border.left + server.width () + border.right;
+		height = border.top  + server.height () + border.bottom;
 
 		x = left + width  * (rKeys[i].dx + 1) / 2;
 		y = top  + height * (rKeys[i].dy + 1) / 2;
@@ -579,7 +741,7 @@ ResizeScreen::handleMotionEvent (int xRoot, int yRoot)
     if (grabIndex)
     {
 	BoxRec box;
-	int    wi, he;                  /* size of window contents */
+	int    wi, he, cwi, che;        /* size of window contents (c prefix for constrained)*/
 	int    wX, wY, wWidth, wHeight; /* rect. for window contents+borders */
 	int    workAreaSnapDistance = 15;
 
@@ -641,16 +803,16 @@ ResizeScreen::handleMotionEvent (int xRoot, int yRoot)
 
 		if (mask & ResizeRightMask)
 			pointerAdjustX = server.x () + server.width () +
-					 w->input ().right - xRoot;
+					 w->border ().right - xRoot;
 		else if (mask & ResizeLeftMask)
-			pointerAdjustX = server.x () - w->input ().left -
+			pointerAdjustX = server.x () - w->border ().left -
 					 xRoot;
 
 		if (mask & ResizeDownMask)
 			pointerAdjustY = server.y () + server.height () +
-					 w->input ().bottom - yRoot;
+					 w->border ().bottom - yRoot;
 		else if (mask & ResizeUpMask)
-			pointerAdjustY = server.y () - w->input ().top - yRoot;
+			pointerAdjustY = server.y () - w->border ().top - yRoot;
 
 		screen->warpPointer (pointerAdjustX, pointerAdjustY);
 
@@ -663,8 +825,52 @@ ResizeScreen::handleMotionEvent (int xRoot, int yRoot)
 	    /* only accumulate pointer movement if a mask is
 	       already set as we don't have a use for the
 	       difference information otherwise */
-	    pointerDx += xRoot - lastPointerX;
-	    pointerDy += yRoot - lastPointerY;
+
+	    if (centered || optionGetResizeFromCenter ())
+	    {
+		pointerDx += (xRoot - lastPointerX) * 2;
+		pointerDy += (yRoot - lastPointerY) * 2;
+	    }
+	    else
+	    {
+		pointerDx += xRoot - lastPointerX;
+		pointerDy += yRoot - lastPointerY;
+	    }
+	    
+	    /* If we hit the edge of the screen while resizing
+	     * the window and the adjacent window edge has not hit
+	     * the edge of the screen, then accumulate pointer motion
+	     * in the opposite direction. (So the apparant x / y
+	     * mixup here is intentional)
+	     */
+
+	    if (isConstrained)
+	    {
+		if (mask == ResizeLeftMask)
+		{
+		    if (xRoot == 0 &&
+			geometry.x - w->border ().left > grabWindowWorkArea->left ())
+			pointerDx += abs (yRoot - lastPointerY) * -1;
+		}
+		else if (mask == ResizeRightMask)
+		{
+		    if (xRoot == screen->width () -1 &&
+			geometry.x + geometry.width + w->border ().right < grabWindowWorkArea->right ())
+			pointerDx += abs (yRoot - lastPointerY);
+		}
+		if (mask == ResizeUpMask)
+		{
+		    if (yRoot == 0 &&
+			geometry.y - w->border ().top > grabWindowWorkArea->top ())
+			pointerDy += abs (xRoot - lastPointerX) * -1;
+		}
+		else if (mask == ResizeDownMask)
+		{
+		    if (yRoot == screen->height () -1 &&
+			geometry.y + geometry.height + w->border ().bottom < grabWindowWorkArea->bottom ())
+			pointerDx += abs (yRoot - lastPointerY);
+		}
+	    }
 	}
 
 	if (mask & ResizeLeftMask)
@@ -683,23 +889,98 @@ ResizeScreen::handleMotionEvent (int xRoot, int yRoot)
 	if (w->state () & CompWindowStateMaximizedHorzMask)
 	    wi = w->serverGeometry ().width ();
 
-	w->constrainNewWindowSize (wi, he, &wi, &he);
+	cwi = wi;
+	che = he;
+
+	if (w->constrainNewWindowSize (wi, he, &cwi, &che) &&
+	    mode != ResizeOptions::ModeNormal)
+	{
+	    Box box;
+
+	    /* Also, damage relevant paint rectangles */
+	    if (mode == ResizeOptions::ModeRectangle ||
+	        mode == ResizeOptions::ModeOutline)
+		getPaintRectangle (&box);
+	    else if (mode == ResizeOptions::ModeStretch)
+		getStretchRectangle (&box);
+
+	    damageRectangle (&box);
+	}
+
+	/* constrain to work area */
+	if (offWorkAreaConstrained)
+	{
+	    if (mask & ResizeUpMask)
+	    {
+		int decorTop = savedGeometry.y + savedGeometry.height -
+		    (che + w->border ().top);
+
+		if (grabWindowWorkArea->y () > decorTop)
+		    che -= grabWindowWorkArea->y () - decorTop;
+	    }
+	    if (mask & ResizeDownMask)
+	    {
+		int decorBottom = savedGeometry.y + che + w->border ().bottom;
+
+		if (decorBottom >
+		    grabWindowWorkArea->y () + grabWindowWorkArea->height ())
+		    che -= decorBottom - (grabWindowWorkArea->y () +
+					  grabWindowWorkArea->height ());
+	    }
+	    if (mask & ResizeLeftMask)
+	    {
+		int decorLeft = savedGeometry.x + savedGeometry.width -
+		    (cwi + w->border ().left);
+
+		if (grabWindowWorkArea->x () > decorLeft)
+		    cwi -= grabWindowWorkArea->x () - decorLeft;
+	    }
+	    if (mask & ResizeRightMask)
+	    {
+		int decorRight = savedGeometry.x + cwi + w->border ().right;
+
+		if (decorRight >
+		    grabWindowWorkArea->x () + grabWindowWorkArea->width ())
+		    cwi -= decorRight - (grabWindowWorkArea->x () +
+				         grabWindowWorkArea->width ());
+	    }
+	}
+
+	wi = cwi;
+	he = che;
 
 	/* compute rect. for window + borders */
-	wWidth  = wi + w->input ().left + w->input ().right;
-	wHeight = he + w->input ().top + w->input ().bottom;
+	wWidth  = wi + w->border ().left + w->border ().right;
+	wHeight = he + w->border ().top + w->border ().bottom;
 
-	if (mask & ResizeLeftMask)
-	    wX = savedGeometry.x + savedGeometry.width -
-		 (wi + w->input ().left);
-	else
-	    wX = savedGeometry.x - w->input ().left;
+	if (centered || optionGetResizeFromCenter ())
+	{
+	    if (mask & ResizeLeftMask)
+		wX = geometry.x + geometry.width -
+		     (wi + w->border ().left);
+	    else
+		wX = geometry.x - w->border ().left;
 
-	if (mask & ResizeUpMask)
-	    wY = savedGeometry.y + savedGeometry.height -
-		 (he + w->input ().top);
+	    if (mask & ResizeUpMask)
+		wY = geometry.y + geometry.height -
+		     (he + w->border ().top);
+	    else
+		wY = geometry.y - w->border ().top;
+	}
 	else
-	    wY = savedGeometry.y - w->input ().top;
+	{
+	    if (mask & ResizeLeftMask)
+		wX = savedGeometry.x + savedGeometry.width -
+		     (wi + w->border ().left);
+	    else
+		wX = savedGeometry.x - w->border ().left;
+
+	    if (mask & ResizeUpMask)
+		wY = savedGeometry.y + savedGeometry.height -
+		     (he + w->border ().top);
+	    else
+		wY = savedGeometry.y - w->border ().top;
+	}
 
 	/* Check if resized edge(s) are near output work-area boundaries */
 	foreach (CompOutput &output, ::screen->outputDevs ())
@@ -769,12 +1050,12 @@ ResizeScreen::handleMotionEvent (int xRoot, int yRoot)
 	    /* rect. for a minimal height window + borders
 	       (used for the constraining in X axis) */
 	    int minimalInputHeight = minHeight +
-				     w->input ().top + w->input ().bottom;
+				     w->border ().top + w->border ().bottom;
 
 	    /* small hot-spot square (on window's corner or edge) that is to be
 	       constrained to the combined output work-area region */
 	    int x, y;
-	    int width = w->input ().top; /* square size = title bar height */
+	    int width = w->border ().top; /* square size = title bar height */
 	    int height = width;
 	    bool status; /* whether or not hot-spot is in the region */
 
@@ -798,8 +1079,8 @@ ResizeScreen::handleMotionEvent (int xRoot, int yRoot)
 	    /* only constrain movement if previous position was valid */
 	    if (inRegionStatus)
 	    {
-		bool xStatus;
-		int yForXResize;
+		bool xStatus = false;
+		int yForXResize = y;
 		int nx = x;
 		int nw = wi;
 		int nh = he;
@@ -922,11 +1203,20 @@ ResizeScreen::handleMotionEvent (int xRoot, int yRoot)
 	    damageRectangle (&box);
 	}
 
-	if (mask & ResizeLeftMask)
-	    geometry.x -= wi - geometry.width;
-
-	if (mask & ResizeUpMask)
-	    geometry.y -= he - geometry.height;
+	if (centered || optionGetResizeFromCenter ())
+	{
+	    if ((mask & ResizeLeftMask) || (mask & ResizeRightMask))
+		geometry.x -= ((wi - geometry.width) / 2);
+	    if ((mask & ResizeUpMask) || (mask & ResizeDownMask))
+		geometry.y -= ((he - geometry.height) / 2);
+	}
+	else
+	{
+	    if (mask & ResizeLeftMask)
+		geometry.x -= wi - geometry.width;
+	    if (mask & ResizeUpMask)
+		geometry.y -= he - geometry.height;
+	}
 
 	geometry.width  = wi;
 	geometry.height = he;
@@ -969,7 +1259,7 @@ ResizeScreen::handleEvent (XEvent *event)
 			CompAction *action = &optionGetInitiateButton ();
 
 			resizeTerminate (action, CompAction::StateTermButton,
-					 noOptions);
+					 noOptions ());
 		    }
 		}
 	    }
@@ -1025,17 +1315,9 @@ ResizeScreen::handleEvent (XEvent *event)
 				ResizeDownMask | ResizeLeftMask,
 				ResizeLeftMask,
 			    };
-			    unsigned int mods;
-			    Window	     root, child;
-			    int	     xRoot, yRoot, i;
-
-			    XQueryPointer (screen->dpy (),
-					   screen->root (),
-					   &root, &child, &xRoot, &yRoot,
-					   &i, &i, &mods);
 
 			    /* TODO: not only button 1 */
-			    if (mods & Button1Mask)
+			    if (pointerMods & Button1Mask)
 			    {
 				o.push_back (CompOption ("modifiers",
 					     CompOption::TypeInt));
@@ -1048,7 +1330,7 @@ ResizeScreen::handleEvent (XEvent *event)
 				o.push_back (CompOption ("button",
 					     CompOption::TypeInt));
 
-				o[2].value ().set ((int) mods);
+				o[2].value ().set ((int) pointerMods);
 				o[3].value ().set
 				    ((int) event->xclient.data.l[0]);
 				o[4].value ().set
@@ -1064,7 +1346,7 @@ ResizeScreen::handleEvent (XEvent *event)
 				    CompAction::StateInitButton, o);
 
 				ResizeScreen::get (screen)->
-				    handleMotionEvent (xRoot, yRoot);
+				    handleMotionEvent (pointerX, pointerY);
 			    }
 			}
 		    }
@@ -1074,9 +1356,9 @@ ResizeScreen::handleEvent (XEvent *event)
 		    if (rs->w->id () == event->xclient.window)
 		    {
 			resizeTerminate (&optionGetInitiateButton (),
-					 CompAction::StateCancel, noOptions);
+					 CompAction::StateCancel, noOptions ());
 			resizeTerminate (&optionGetInitiateKey (),
-					 CompAction::StateCancel, noOptions);
+					 CompAction::StateCancel, noOptions ());
 		    }
 		}
 	    }
@@ -1084,20 +1366,133 @@ ResizeScreen::handleEvent (XEvent *event)
 	case DestroyNotify:
 	    if (w && w->id () == event->xdestroywindow.window)
 	    {
-		resizeTerminate (&optionGetInitiateButton (), 0, noOptions);
-		resizeTerminate (&optionGetInitiateKey (), 0, noOptions);
+		resizeTerminate (&optionGetInitiateButton (), 0, noOptions ());
+		resizeTerminate (&optionGetInitiateKey (), 0, noOptions ());
 	    }
 	    break;
 	case UnmapNotify:
 	    if (w && w->id () == event->xunmap.window)
 	    {
-		resizeTerminate (&optionGetInitiateButton (), 0, noOptions);
-		resizeTerminate (&optionGetInitiateKey (), 0, noOptions);
+		resizeTerminate (&optionGetInitiateButton (), 0, noOptions ());
+		resizeTerminate (&optionGetInitiateKey (), 0, noOptions ());
 	    }
 	default:
 	    break;
     }
 
+    if (event->type == screen->xkbEvent ())
+    {
+	XkbAnyEvent *xkbEvent = (XkbAnyEvent *) event;
+
+	if (xkbEvent->xkb_type == XkbStateNotify)
+	{
+	    XkbStateNotifyEvent *stateEvent = (XkbStateNotifyEvent *) event;
+
+	    /* Check if we need to change to outline mode */
+
+	    unsigned int mods = 0xffffffff;
+	    bool	 modifierMode = false;
+	    int		 oldMode = mode;
+
+	    if (outlineMask)
+		mods = outlineMask;
+
+	    if ((stateEvent->mods & mods) == mods)
+	    {
+		modifierMode = true;
+		mode = ResizeOptions::ModeOutline;
+	    }
+
+	    mods = 0xffffffff;
+	    if (rectangleMask)
+		mods = rectangleMask;
+
+	    if ((stateEvent->mods & mods) == mods)
+	    {
+		modifierMode = true;
+		mode = ResizeOptions::ModeRectangle;
+	    }
+
+	    mods = 0xffffffff;
+	    if (stretchMask)
+		mods = stretchMask;
+
+	    if ((stateEvent->mods & mods) == mods)
+	    {
+		modifierMode = true;
+		mode = ResizeOptions::ModeStretch;
+	    }
+
+	    mods = 0xffffffff;
+	    if (centeredMask)
+		mods = centeredMask;
+
+	    /* No modifier mode set, check match options */
+	    if (w)
+	    {
+		if (optionGetNormalMatch ().evaluate (w))
+		{
+		    modifierMode = true;
+		    mode = ResizeOptions::ModeNormal;
+		}
+
+		if (optionGetOutlineMatch ().evaluate (w))
+		{
+		    modifierMode = true;
+		    mode = ResizeOptions::ModeOutline;
+		}
+
+		if (optionGetRectangleMatch ().evaluate (w))
+		{
+		    modifierMode = true;
+		    mode = ResizeOptions::ModeRectangle;
+		}
+
+		if (optionGetStretchMatch ().evaluate (w))
+		{
+		    modifierMode = true;
+		    mode = ResizeOptions::ModeStretch;
+		}
+	    }
+
+	    if (!modifierMode)
+		mode = optionGetMode ();
+
+	    if (w && oldMode != mode)
+	    {
+		Box box;
+
+		getStretchRectangle (&box);
+		damageRectangle (&box);
+		getPaintRectangle (&box);
+		damageRectangle (&box);
+
+		box.x1 = w->outputRect ().x ();
+		box.y1 = w->outputRect ().y ();
+		box.x2 = box.x1 + w->outputRect ().width ();
+		box.y2 = box.y1 + w->outputRect ().height ();
+
+		damageRectangle (&box);
+	    }
+
+	    if ((stateEvent->mods & mods) == mods)
+	    {
+		centered = true;
+	    }
+	    else if (w)
+	    {
+		if (!optionGetResizeFromCenterMatch ().evaluate (w))
+		    centered = false;
+		else
+		    centered = true;
+	    }
+	    else
+	    {
+		centered = false;
+	    }
+	}
+    }
+
     screen->handleEvent (event);
 
     if (event->type == screen->syncEvent () + XSyncAlarmNotify)
@@ -1130,8 +1525,20 @@ ResizeScreen::glPaintRectangle (const GLScreenPaintAttrib &sAttrib,
 				unsigned short            *borderColor,
 				unsigned short            *fillColor)
 {
-    BoxRec   box;
-    GLMatrix sTransform (transform);
+    BoxRec   	   box;
+    GLMatrix 	   sTransform (transform);
+    GLint    	   origSrc, origDst;
+    float_t	   fc[4], bc[4];
+
+    glGetIntegerv (GL_BLEND_SRC, &origSrc);
+    glGetIntegerv (GL_BLEND_DST, &origDst);
+
+    /* Premultiply the alpha values */
+    
+    bc[3] = (float) borderColor[3] / (float) 65535.0f;
+    bc[0] = ((float) borderColor[0] / 65535.0f) * bc[3];
+    bc[1] = ((float) borderColor[1] / 65535.0f) * bc[3];
+    bc[2] = ((float) borderColor[2] / 65535.0f) * bc[3];
 
     getPaintRectangle (&box);
 
@@ -1143,16 +1550,22 @@ ResizeScreen::glPaintRectangle (const GLScreenPaintAttrib &sAttrib,
 
     glDisableClientState (GL_TEXTURE_COORD_ARRAY);
     glEnable (GL_BLEND);
+    glBlendFunc (GL_ONE, GL_ONE_MINUS_SRC_ALPHA);
 
     /* fill rectangle */
     if (fillColor)
     {
-	glColor4usv (fillColor);
+	fc[3] = (float) fillColor[3] / (float) 65535.0f;
+	fc[0] = ((float) fillColor[0] / 65535.0f) * fc[3];
+	fc[1] = ((float) fillColor[1] / 65535.0f) * fc[3];
+	fc[2] = ((float) fillColor[2] / 65535.0f) * fc[3];
+
+	glColor4f (fc[0], fc[1], fc[2], fc[3]);
 	glRecti (box.x1, box.y2, box.x2, box.y1);
     }
 
     /* draw outline */
-    glColor4usv (borderColor);
+    glColor4f (bc[0], bc[1], bc[2], bc[3]);
     glLineWidth (2.0);
     glBegin (GL_LINE_LOOP);
     glVertex2i (box.x1, box.y1);
@@ -1239,8 +1652,8 @@ ResizeWindow::glPaint (const GLWindowPaintAttrib &attrib,
 	x = window->geometry (). x ();
 	y = window->geometry (). y ();
 
-	xOrigin = x - window->input ().left;
-	yOrigin = y - window->input ().top;
+	xOrigin = x - window->border ().left;
+	yOrigin = y - window->border ().top;
 
 	wTransform.translate (xOrigin, yOrigin, 0.0f);
 	wTransform.scale (xScale, yScale, 1.0f);
@@ -1284,17 +1697,80 @@ ResizeWindow::damageRect (bool initial, const CompRect &rect)
     return status;
 }
 
+/* We have to make some assumptions here in order to do this neatly,
+ * see build/generated/resize_options.h for more info */
+
+#define ResizeModeShiftMask (1 << 0)
+#define ResizeModeAltMask (1 << 1)
+#define ResizeModeControlMask (1 << 2)
+#define ResizeModeMetaMask (1 << 3)
+
+void
+ResizeScreen::resizeMaskValueToKeyMask (int	   valueMask,
+					int	   *mask)
+{
+    if (valueMask & ResizeModeShiftMask)
+	*mask |= ShiftMask;
+    if (valueMask & ResizeModeAltMask)
+	*mask |= CompAltMask;
+    if (valueMask & ResizeModeControlMask)
+	*mask |= ControlMask;
+    if (valueMask & ResizeModeMetaMask)
+	*mask |= CompMetaMask;
+}
+
+void
+ResizeScreen::optionChanged (CompOption		    *option,
+			     ResizeOptions::Options num)
+{
+    int *mask = NULL;
+    int valueMask = 0;
+
+    switch (num)
+    {
+	case ResizeOptions::OutlineModifier:
+	    mask = &outlineMask;
+	    valueMask = optionGetOutlineModifierMask ();
+	    break;
+	case ResizeOptions::RectangleModifier:
+	    mask = &rectangleMask;
+	    valueMask = optionGetRectangleModifierMask ();
+	    break;
+	case ResizeOptions::StretchModifier:
+	    mask = &stretchMask;
+	    valueMask = optionGetStretchModifierMask ();
+	    break;
+        case ResizeOptions::CenteredModifier:
+	    mask = &centeredMask;
+	    valueMask = optionGetCenteredModifierMask ();
+	    break;
+	default:
+	    break;
+    }
+
+    if (mask)
+	resizeMaskValueToKeyMask (valueMask, mask);
+}
+
 ResizeScreen::ResizeScreen (CompScreen *s) :
     PluginClassHandler<ResizeScreen,CompScreen> (s),
     gScreen (GLScreen::get (s)),
     cScreen (CompositeScreen::get (s)),
     w (NULL),
+    outlineMask (0),
+    rectangleMask (0),
+    stretchMask (0),
+    centeredMask (0),
     releaseButton (0),
-    isConstrained (false)
+    isConstrained (false),
+    offWorkAreaConstrained (true),
+    grabWindowWorkArea ()
 {
     CompOption::Vector atomTemplate;
     Display *dpy = s->dpy ();
-    
+    ResizeOptions::ChangeNotify notify =
+	       boost::bind (&ResizeScreen::optionChanged, this, _1, _2);
+
     atomTemplate.resize (4);
 
     for (int i = 0; i < 4; i++)
@@ -1330,27 +1806,21 @@ ResizeScreen::ResizeScreen (CompScreen *s) :
     cursor[2] = upCursor;
     cursor[3] = downCursor;
 
-    optionSetInitiateNormalKeyInitiate (boost::bind
-					 (resizeInitiate, _1, _2, _3,
-					  ResizeOptions::ModeNormal));
-    optionSetInitiateNormalKeyTerminate (resizeTerminate);
-    optionSetInitiateOutlineKeyInitiate (boost::bind
-					 (resizeInitiate, _1, _2, _3,
-					  ResizeOptions::ModeOutline));
-    optionSetInitiateOutlineKeyTerminate (resizeTerminate);
-    optionSetInitiateRectangleKeyInitiate (boost::bind
-					   (resizeInitiate, _1, _2, _3,
-					    ResizeOptions::ModeRectangle));
-    optionSetInitiateRectangleKeyTerminate (resizeTerminate);
-    optionSetInitiateStretchKeyInitiate (boost::bind
-					 (resizeInitiate, _1, _2, _3,
-					  ResizeOptions::ModeStretch));
-    optionSetInitiateStretchKeyTerminate (resizeTerminate);
     optionSetInitiateKeyInitiate (resizeInitiateDefaultMode);
     optionSetInitiateKeyTerminate (resizeTerminate);
     optionSetInitiateButtonInitiate (resizeInitiateDefaultMode);
     optionSetInitiateButtonTerminate (resizeTerminate);
 
+    optionSetOutlineModifierNotify (notify);
+    optionSetRectangleModifierNotify (notify);
+    optionSetStretchModifierNotify (notify);
+    optionSetCenteredModifierNotify (notify);
+
+    resizeMaskValueToKeyMask (optionGetOutlineModifierMask (), &outlineMask);
+    resizeMaskValueToKeyMask (optionGetRectangleModifierMask (), &rectangleMask);
+    resizeMaskValueToKeyMask (optionGetStretchModifierMask (), &stretchMask);
+    resizeMaskValueToKeyMask (optionGetCenteredModifierMask (), &centeredMask);
+
     ScreenInterface::setHandler (s);
 
     if (gScreen)
