@@ -90,7 +90,7 @@ PrivateScaleWindow::isScaleWin () const
 		return false;
 	    break;
 	case ScaleTypeOutput:
-	    if ((unsigned int) window->outputDevice () != 
+	    if ((unsigned int) window->outputDevice () !=
 	    		       (unsigned int) screen->currentOutputDev ().id ())
 		return false;
 	default:
@@ -130,7 +130,7 @@ ScaleWindow::scalePaintDecoration (const GLWindowPaintAttrib& attrib,
 				   const CompRegion&          region,
 				   unsigned int               mask)
 {
-    WRAPABLE_HND_FUNC (0, scalePaintDecoration, attrib, transform, region, mask)
+    WRAPABLE_HND_FUNCTN (scalePaintDecoration, attrib, transform, region, mask)
 
     if (priv->spScreen->optionGetOverlayIcon () != ScaleOptions::OverlayIconNone)
     {
@@ -147,7 +147,6 @@ ScaleWindow::scalePaintDecoration (const GLWindowPaintAttrib& attrib,
 	    float  x, y;
 	    int    width, height;
 	    int    scaledWinWidth, scaledWinHeight;
-	    float  ds;
 
 	    scaledWinWidth  = priv->window->width () * priv->scale;
 	    scaledWinHeight = priv->window->height () * priv->scale;
@@ -194,6 +193,7 @@ ScaleWindow::scalePaintDecoration (const GLWindowPaintAttrib& attrib,
 	    if (priv->delta)
 	    {
 		float o;
+		float ds;
 
 		ds = fabs (priv->tx) +
 		     fabs (priv->ty) +
@@ -257,10 +257,75 @@ ScaleWindowInterface::setScaledPaintAttributes (GLWindowPaintAttrib& attrib)
 bool
 ScaleWindow::setScaledPaintAttributes (GLWindowPaintAttrib& attrib)
 {
-    WRAPABLE_HND_FUNC_RETURN (1, bool, setScaledPaintAttributes, attrib)
+    WRAPABLE_HND_FUNCTN_RETURN (bool, setScaledPaintAttributes, attrib)
 
     bool drawScaled = false;
 
+    /* Windows that wouldn't be visible before and after entering
+     * scale mode (because some plugin modified CompWindow::focus)
+     * should be faded in and out */
+    if (window->state () & CompWindowStateHiddenMask)
+    {
+	GLfloat factor = 0;
+	GLfloat targetX, targetY, targetScale;
+	GLfloat scaleFactor, xFactor, yFactor, divFactor = 3.0f;
+
+	if (priv->slot)
+	{
+	    targetX = priv->slot->x ();
+	    targetY = priv->slot->y ();
+	    targetScale = priv->slot->scale;
+	}
+	else
+	{
+	    targetX = priv->lastTargetX;
+	    targetY = priv->lastTargetY;
+	    targetScale = priv->lastTargetScale;
+	}
+
+	/* Don't FDIV0 */
+	if (targetScale - priv->scale == 0.0f)
+	{
+	    divFactor -= 1.0f;
+	    scaleFactor = 1.0f;
+	}
+	else
+	    scaleFactor = (1.0f - priv->scale) / (1.0f - targetScale);
+
+	if (targetX - ((float) window->x () + priv->tx) == 0.0f)
+	{
+	    divFactor -= 1.0f;
+	    xFactor = 1.0f;
+	}
+	else
+	{
+	    float distActual = fabsf (window->x () - ((float) window->x () + priv->tx));
+	    float distTarget = fabsf (window->x () - targetX);
+
+	    xFactor = distActual / distTarget;
+	}
+
+	if (targetY - ((float) window->y () + priv->ty) == 0.0f)
+	{
+	    divFactor -= 1.0f;
+	    yFactor = 1.0f;
+	}
+	else
+	{
+	    float distActual = fabsf (window->y () - ((float) window->y () + priv->ty));
+	    float distTarget = fabsf (window->y () - targetY);
+
+	    yFactor = distActual / distTarget;
+	}
+
+	if (divFactor)
+	    factor = (scaleFactor + xFactor + yFactor) / divFactor;
+	else
+	    factor = 1.0f;
+
+	attrib.opacity *= factor;
+    }
+
     if (priv->adjust || priv->slot)
     {
 	if (priv->window->id ()     != priv->spScreen->selectedWindow &&
@@ -412,64 +477,22 @@ PrivateScaleScreen::layoutSlotsForArea (const CompRect& workArea,
 SlotArea::vector
 PrivateScaleScreen::getSlotAreas ()
 {
-    unsigned int       i = 0;
-    CompRect           workArea;
-    std::vector<float> size;
-    float              sizePerWindow, sum = 0.0f;
-    int                left;
-    SlotArea::vector   slotAreas;
+    SlotArea::vector slotAreas;
 
     slotAreas.resize (screen->outputDevs ().size ());
-    size.resize (screen->outputDevs ().size ());
-
-    left = windows.size ();
 
+    unsigned int i = 0;
     foreach (CompOutput &o, screen->outputDevs ())
     {
-	/* determine the size of the workarea for each output device */
-	workArea = CompRect (o.workArea ());
-
-	size[i] = workArea.width () * workArea.height ();
-	sum += size[i];
-
 	slotAreas[i].nWindows = 0;
-	slotAreas[i].workArea = workArea;
-
-	i++;
-    }
-
-    /* calculate size available for each window */
-    sizePerWindow = sum / windows.size ();
-
-    for (i = 0; i < screen->outputDevs ().size () && left; i++)
-    {
-	/* fill the areas with windows */
-	int nw = floor (size[i] / sizePerWindow);
-
-	nw = MIN (nw, left);
-	size[i] -= nw * sizePerWindow;
-	slotAreas[i].nWindows = nw;
-	left -= nw;
-    }
-
-    /* add left windows to output devices with the biggest free space */
-    while (left > 0)
-    {
-	int   num = 0;
-	float big = 0;
-
-	for (i = 0; i < screen->outputDevs ().size (); i++)
+	foreach (ScaleWindow *window, windows)
 	{
-	    if (size[i] > big)
-	    {
-		num = i;
-		big = size[i];
-	    }
+	    CompWindow *cw = window->priv->window;
+	    if (cw->outputDevice () == (int) o.id ())
+		slotAreas[i].nWindows++;
 	}
 
-	size[num] -= sizePerWindow;
-	slotAreas[num].nWindows++;
-	left--;
+	slotAreas[i++].workArea = o.workArea ();
     }
 
     return slotAreas;
@@ -535,8 +558,8 @@ PrivateScaleScreen::findBestSlots ()
 		sx = (slots[i].x2 () + slots[i].x1 ()) / 2;
 		sy = (slots[i].y2 () + slots[i].y1 ()) / 2;
 
-		cx = w->serverX () + w->width () / 2;
-		cy = w->serverY () + w->height () / 2;
+		cx = (w->serverX () - (w->defaultViewport ().x () - screen->vp ().x ()) * screen->width ()) + w->width () / 2;
+		cy = (w->serverY () - (w->defaultViewport ().y () - screen->vp ().y ()) * screen->height ()) + w->height () / 2;
 
 		cx -= sx;
 		cy -= sy;
@@ -572,6 +595,9 @@ PrivateScaleScreen::fillInWindows ()
 
 	    sw->priv->slot = &slots[sw->priv->sid];
 
+	    /* Auxilary items reparented into windows are clickable so we want to care about
+	     * them when calculating the slot size */
+
 	    width  = w->width ()  + w->input ().left + w->input ().right;
 	    height = w->height () + w->input ().top  + w->input ().bottom;
 
@@ -588,9 +614,7 @@ PrivateScaleScreen::fillInWindows ()
 	    cx += w->input ().left * sw->priv->slot->scale;
 	    cy += w->input ().top  * sw->priv->slot->scale;
 
-	    sw->priv->slot->setGeometry (cx - sx / 2, cy - sy / 2,
-					 (cx - sx / 2) + (cx + sx / 2),
-					 (cy - sy / 2) + (cy + sy / 2));
+	    sw->priv->slot->setGeometry (cx - sx / 2, cy - sy / 2, sx, sy);
 
 	    sw->priv->slot->filled = true;
 
@@ -610,7 +634,7 @@ ScaleScreenInterface::layoutSlotsAndAssignWindows ()
 bool
 ScaleScreen::layoutSlotsAndAssignWindows ()
 {
-    WRAPABLE_HND_FUNC_RETURN (0, bool, layoutSlotsAndAssignWindows)
+    WRAPABLE_HND_FUNCTN_RETURN (bool, layoutSlotsAndAssignWindows)
 
     /* create a grid of slots */
     priv->layoutSlots ();
@@ -660,30 +684,61 @@ ScaleScreen::getWindows () const
 bool
 PrivateScaleScreen::layoutThumbs ()
 {
-    windows.clear ();
+    bool ret = false;
+    std::map <ScaleWindow *, ScaleSlot> slotWindows;
+    CompWindowList          allWindows;
 
-    /* add windows scale list, top most window first */
-    foreach (CompWindow *w, screen->windows ())
+    for (int i = 0; i < screen->vpSize ().height (); i++)
     {
-	SCALE_WINDOW (w);
+	for (int j = 0; j < screen->vpSize ().width (); j++)
+	{
+	    windows.clear ();
+	    slots.clear ();
 
-	if (sw->priv->slot)
-	    sw->priv->adjust = true;
+	    /* add windows scale list, top most window first */
+	    foreach (CompWindow *w, screen->windows ())
+	    {
+		SCALE_WINDOW (w);
 
-	sw->priv->slot = NULL;
+		if (w->defaultViewport () != CompPoint (j, i))
+		    continue;
 
-	if (!sw->priv->isScaleWin ())
-	    continue;
+		if (sw->priv->slot)
+		    sw->priv->adjust = true;
 
-	windows.push_back (sw);
+		sw->priv->slot = NULL;
+
+		if (!sw->priv->isScaleWin ())
+		    continue;
+
+		windows.push_back (sw);
+	    }
+
+	    if (!windows.empty ())
+	    {
+		slots.resize (windows.size ());
+		ret |= ScaleScreen::get (screen)->layoutSlotsAndAssignWindows ();
+
+		foreach (ScaleWindow *sw, windows)
+		    slotWindows[sw] = *sw->priv->slot;
+	    }
+	}
     }
 
-    if (windows.empty ())
-	return false;
+    slots.clear ();
+    windows.clear ();
 
-    slots.resize (windows.size ());
+    for (std::map<ScaleWindow *, ScaleSlot>::iterator it = slotWindows.begin ();
+	 it != slotWindows.end (); it++)
+    {
+	slots.push_back (it->second);
+	windows.push_back (it->first);
+	it->first->priv->slot = &slots.back ();
+	it->first->priv->slot->setX (it->first->priv->slot->x () + (it->first->priv->window->defaultViewport ().x () - screen->vp ().x ()) * screen->width ());
+	it->first->priv->slot->setY (it->first->priv->slot->y () + (it->first->priv->window->defaultViewport ().y () - screen->vp ().y ()) * screen->height ());
+    }
 
-    return ScaleScreen::get (screen)->layoutSlotsAndAssignWindows ();
+    return ret;
 }
 
 bool
@@ -909,9 +964,8 @@ PrivateScaleScreen::scaleTerminate (CompAction         *action,
     SCALE_SCREEN (screen);
 
     Window xid;
-
-    action->setState (action->state () & ~(CompAction::StateTermKey |
-					   CompAction::StateTermButton));
+    int    selectX = CompOption::getIntOptionNamed (options, "select_x", -1);
+    int    selectY = CompOption::getIntOptionNamed (options, "select_y", -1);
 
     if (ss->priv->actionShouldToggle (action, state))
 	return false;
@@ -923,6 +977,13 @@ PrivateScaleScreen::scaleTerminate (CompAction         *action,
     if (!ss->priv->grab)
 	return false;
 
+    if (selectX != -1 &&
+	selectY != -1)
+    {
+	if (!ss->priv->selectWindowAt (selectX, selectY, true))
+	    return false;
+    }
+
     if (ss->priv->grabIndex)
     {
 	::screen->removeGrab (ss->priv->grabIndex, 0);
@@ -942,9 +1003,18 @@ PrivateScaleScreen::scaleTerminate (CompAction         *action,
 
 	    if (sw->priv->slot)
 	    {
+		sw->priv->lastTargetScale = sw->priv->slot->scale;
+		sw->priv->lastTargetX     = sw->priv->slot->x ();
+		sw->priv->lastTargetY     = sw->priv->slot->y ();
 		sw->priv->slot   = NULL;
 		sw->priv->adjust = true;
 	    }
+	    else
+	    {
+		sw->priv->lastTargetScale = 1.0f;
+		sw->priv->lastTargetX = w->x ();
+		sw->priv->lastTargetY = w->y ();
+	    }
 	}
 
 	if (state & CompAction::StateCancel)
@@ -969,8 +1039,15 @@ PrivateScaleScreen::scaleTerminate (CompAction         *action,
 	ss->priv->cScreen->damageScreen ();
     }
 
+    if (state & CompAction::StateInitKey)
+	action->setState (action->state () | CompAction::StateTermKey);
+
     ss->priv->lastActiveNum = 0;
 
+    if (selectX != -1 &&
+	selectY != -1)
+	return true;
+
     return false;
 }
 
@@ -1038,19 +1115,20 @@ PrivateScaleScreen::scaleInitiate (CompAction         *action,
     {
 	SCALE_SCREEN (::screen);
 
-	if (ss->priv->state != ScaleScreen::Wait &&
-	    ss->priv->state != ScaleScreen::Out)
-	{
-	    ss->priv->type = type;
-	    return ss->priv->scaleInitiateCommon (action, state, options);
-	}
-	else if (ss->priv->actionShouldToggle (action, state))
+	if (ss->priv->actionShouldToggle (action, state) &&
+	    (ss->priv->state == ScaleScreen::Wait ||
+	     ss->priv->state == ScaleScreen::Out))
 	{
 	    if (ss->priv->type == type)
 		return scaleTerminate (action,
 				       CompAction::StateCancel,
 				       options);
 	}
+	else
+	{
+	    ss->priv->type = type;
+	    return ss->priv->scaleInitiateCommon (action, state, options);
+	}
     }
 
     return false;
@@ -1061,7 +1139,9 @@ PrivateScaleScreen::scaleInitiateCommon (CompAction         *action,
 					 CompAction::State  state,
 					 CompOption::Vector &options)
 {
-    if (screen->otherGrabExist ("scale", NULL))
+    int noAutoGrab = CompOption::getIntOptionNamed (options, "no_auto_grab", 0);
+
+    if (screen->otherGrabExist ("scale", NULL) && !noAutoGrab)
 	return false;
 
     match = CompOption::getMatchOptionNamed (options, "match",
@@ -1074,16 +1154,22 @@ PrivateScaleScreen::scaleInitiateCommon (CompAction         *action,
     if (!layoutThumbs ())
 	return false;
 
-    if (state & CompAction::StateInitEdgeDnd)
+    /* Another plugin may be using us externally */
+    grab = noAutoGrab;
+
+    if (!grab)
     {
-	if (ensureDndRedirectWindow ())
-	    grab = true;
-    }
-    else if (!grabIndex)
-    {
-	grabIndex = screen->pushGrab (cursor, "scale");
-	if (grabIndex)
-	    grab = true;
+	if (state & CompAction::StateInitEdgeDnd)
+	{
+	    if (ensureDndRedirectWindow ())
+		grab = true;
+	}
+	else if (!grabIndex)
+	{
+	    grabIndex = screen->pushGrab (cursor, "scale");
+	    if (grabIndex)
+		grab = true;
+	}
     }
 
     if (grab)
@@ -1133,7 +1219,7 @@ ScaleWindowInterface::scaleSelectWindow ()
 void
 ScaleWindow::scaleSelectWindow ()
 {
-    WRAPABLE_HND_FUNC (2, scaleSelectWindow)
+    WRAPABLE_HND_FUNCTN (scaleSelectWindow)
 
     if (priv->spScreen->selectedWindow != priv->window->id ())
     {
@@ -1178,7 +1264,7 @@ ScaleWindow::setSlot (const ScaleSlot &newSlot)
     priv->adjust = true;
 
     if (!priv->slot)
-	return;
+	priv->slot = new ScaleSlot ();
     *priv->slot = newSlot;
 
     /* Trigger the animation to this point */
@@ -1350,9 +1436,8 @@ ScaleScreen::relayoutSlots (const CompMatch& match)
 }
 
 void
-PrivateScaleScreen::windowRemove (Window id)
+PrivateScaleScreen::windowRemove (CompWindow *w)
 {
-    CompWindow *w = screen->findWindow (id);
     if (!w)
 	return;
 
@@ -1421,6 +1506,8 @@ PrivateScaleScreen::hoverTimeout ()
 void
 PrivateScaleScreen::handleEvent (XEvent *event)
 {
+    CompWindow *w = NULL;
+
     switch (event->type) {
 	case KeyPress:
 	    if (screen->root () == event->xkey.root)
@@ -1439,8 +1526,7 @@ PrivateScaleScreen::handleEvent (XEvent *event)
 	    }
 	    break;
 	case ButtonPress:
-	    if (event->xbutton.button == Button1       &&
-		screen->root () == event->xbutton.root &&
+	    if (screen->root () == event->xbutton.root &&
 		grabIndex                              &&
 		state != ScaleScreen::In)
 	    {
@@ -1450,12 +1536,16 @@ PrivateScaleScreen::handleEvent (XEvent *event)
 		o.push_back (CompOption ("root", CompOption::TypeInt));
 		o[0].value ().set ((int) screen->root ());
 
-		if (selectWindowAt (button->x_root, button->y_root, true))
+		/* Button1 terminates scale mode, other buttons can select
+		 * windows */
+		if (selectWindowAt (button->x_root, button->y_root, true) &&
+		    event->xbutton.button == Button1)
 		{
 		    scaleTerminate (&optionGetInitiateEdge (), 0, o);
 		    scaleTerminate (&optionGetInitiateKey (), 0, o);
 		}
-		else if (optionGetShowDesktop ())
+		else if (optionGetShowDesktop () &&
+			 event->xbutton.button == Button1)
 		{
 		    CompPoint pointer (button->x_root, button->y_root);
 		    CompRect  workArea (screen->workArea ());
@@ -1477,7 +1567,7 @@ PrivateScaleScreen::handleEvent (XEvent *event)
 		bool       focus = false;
 		CompOption *o = screen->getOption ("click_to_focus");
 
-		if (o && o->value ().b ())
+		if (o && !o->value ().b ())
 		    focus = true;
 
 		selectWindowAt (event->xmotion.x_root,
@@ -1485,18 +1575,30 @@ PrivateScaleScreen::handleEvent (XEvent *event)
 				focus);
 	    }
 	    break;
+	case DestroyNotify:
+
+	    /* We need to get the CompWindow * for event->xdestroywindow.window
+	     * here because in the ::handleEvent call below that CompWindow's
+	     * id will become "1" so CompScreen::findWindow won't
+	     * be able to find teh window after that
+	     */
+
+	    w = screen->findWindow (event->xdestroywindow.window);
+	    break;
+	case UnmapNotify:
+
+	     w = screen->findWindow (event->xunmap.window);
+	     break;
 	case ClientMessage:
 	    if (event->xclient.message_type == Atoms::xdndPosition)
 	    {
-		CompWindow *w;
-
 		w = screen->findWindow (event->xclient.window);
 		if (w)
 		{
 		    bool       focus = false;
 		    CompOption *o = screen->getOption ("click_to_focus");
 
-		    if (o && o->value ().b ())
+		    if (o && !o->value ().b ())
 			focus = true;
 
 		    if (w->id () == dndTarget)
@@ -1506,10 +1608,7 @@ PrivateScaleScreen::handleEvent (XEvent *event)
 			state != ScaleScreen::In &&
 			w->id () == dndTarget)
 		    {
-			int x = event->xclient.data.l[2] >> 16;
-			int y = event->xclient.data.l[2] & 0xffff;
-
-			ScaleWindow *sw = checkForWindowAt (x, y);
+			ScaleWindow *sw = checkForWindowAt (pointerX, pointerY);
 			if (sw && sw->priv->isScaleWin ())
 			{
 			    int time;
@@ -1518,14 +1617,18 @@ PrivateScaleScreen::handleEvent (XEvent *event)
 
 			    if (hover.active ())
 			    {
-				if (w->id () != selectedWindow)
+				int lastMotion = sqrt (pow (pointerX - lastPointerX, 2) + pow (pointerY - lastPointerY, 2));
+				
+				if (sw->window->id () != selectedWindow || lastMotion > optionGetDndDistance ())
 				    hover.stop ();
 			    }
 
 			    if (!hover.active ())
+			    {
 				hover.start (time, (float) time * 1.2);
+			    }
 
-			    selectWindowAt (x, y, focus);
+			    selectWindowAt (pointerX, pointerY, focus);
 			}
 			else
 			{
@@ -1538,7 +1641,7 @@ PrivateScaleScreen::handleEvent (XEvent *event)
 	    else if (event->xclient.message_type == Atoms::xdndDrop ||
 		     event->xclient.message_type == Atoms::xdndLeave)
 	    {
-		CompWindow *w = screen->findWindow (event->xclient.window);
+		w = screen->findWindow (event->xclient.window);
 		if (w)
 		{
 		    if (grab			 &&
@@ -1560,12 +1663,18 @@ PrivateScaleScreen::handleEvent (XEvent *event)
 
     screen->handleEvent (event);
 
+    /* Only safe to remove the window after all events have been
+     * handled, so that we don't get race conditions on calls
+     * to scale functions */
+
     switch (event->type) {
 	case UnmapNotify:
-	    windowRemove (event->xunmap.window);
+	    if (w)
+		windowRemove (w);
 	    break;
 	case DestroyNotify:
-	    windowRemove (event->xdestroywindow.window);
+	    if (w)
+		windowRemove (w);
 	    break;
     }
 }
@@ -1615,6 +1724,7 @@ ScaleScreen::~ScaleScreen ()
 
 ScaleWindow::ScaleWindow (CompWindow *w) :
     PluginClassHandler<ScaleWindow, CompWindow, COMPIZ_SCALE_ABI> (w),
+    window (w),
     priv (new PrivateScaleWindow (w))
 {
 }
@@ -1733,6 +1843,9 @@ PrivateScaleWindow::PrivateScaleWindow (CompWindow *w) :
     yVelocity (0.0),
     scaleVelocity (0.0),
     scale (1.0),
+    lastTargetScale (1.0f),
+    lastTargetX (w->x ()),
+    lastTargetY (w->y ()),
     tx (0.0),
     ty (0.0),
     delta (1.0),
