@@ -36,14 +36,12 @@
 #include <X11/extensions/Xrandr.h>
 #include <X11/extensions/Xfixes.h>
 
-#include <core/core.h>
 #include <core/atoms.h>
 #include "privatescreen.h"
 #include "privatewindow.h"
+#include "privatestackdebugger.h"
 
-static Window xdndWindow = None;
-static Window edgeWindow = None;
-
+namespace cps = compiz::private_screen;
 
 
 bool
@@ -123,22 +121,65 @@ isInitiateBinding (CompOption	           &option,
 }
 
 static bool
-isTerminateBinding (CompOption	            &option,
-		    CompAction::BindingType type,
-		    CompAction::State       state,
-		    CompAction              **action)
+isBound (CompOption             &option,
+         CompAction::BindingType type,
+         CompAction::State       state,
+         CompAction            **action)
 {
     if (!isCallBackBinding (option, type, state))
 	return false;
 
-    if (option.value ().action ().terminate ().empty ())
-	return false;
-
     *action = &option.value ().action ();
 
     return true;
 }
 
+bool
+cps::EventManager::triggerPress (CompAction         *action,
+                             CompAction::State   state,
+                             CompOption::Vector &arguments)
+{
+    bool actionEventHandled = false;
+
+    if (state == CompAction::StateInitKey && grabsEmpty ())
+    {
+        possibleTap = action;
+    }
+
+    if (!action->initiate ().empty ())
+    {
+	if (action->initiate () (action, state, arguments))
+	    actionEventHandled = true;
+    }
+    else if (!action->terminate ().empty ())
+    {
+	/* Default Initiate implementation for plugins that only
+	   provide a Terminate callback */
+	if (state & CompAction::StateInitKey)
+	    action->setState (action->state () | CompAction::StateTermKey);
+    }
+
+    return actionEventHandled;
+}
+
+bool
+cps::EventManager::triggerRelease (CompAction         *action,
+                               CompAction::State   state,
+                               CompOption::Vector &arguments)
+{
+    if (action == possibleTap)
+    {
+        state |= CompAction::StateTermTapped;
+        possibleTap = NULL;
+    }
+
+    if (!action->terminate ().empty () &&
+        action->terminate () (action, state, arguments))
+        return true;
+
+    return false;
+}
+
 bool
 PrivateScreen::triggerButtonPressBindings (CompOption::Vector &options,
 					   XButtonEvent       *event,
@@ -155,12 +196,12 @@ PrivateScreen::triggerButtonPressBindings (CompOption::Vector &options,
     {
 	unsigned int i;
 
-	if (event->root != root)
+	if (event->root != screen->root())
 	    return false;
 
 	if (event->window != edgeWindow)
 	{
-	    if (grabs.empty () || event->window != root)
+	    if (grabsEmpty () || event->window != screen->root())
 		return false;
 	}
 
@@ -177,8 +218,7 @@ PrivateScreen::triggerButtonPressBindings (CompOption::Vector &options,
 
     foreach (CompOption &option, options)
     {
-	if (isInitiateBinding (option, CompAction::BindingTypeButton, state,
-			       &action))
+	if (isBound (option, CompAction::BindingTypeButton, state, &action))
 	{
 	    if (action->button ().button () == (int) event->button)
 	    {
@@ -186,8 +226,10 @@ PrivateScreen::triggerButtonPressBindings (CompOption::Vector &options,
 		    action->button ().modifiers ());
 
 		if ((bindMods & modMask) == (event->state & modMask))
-		    if (action->initiate () (action, state, arguments))
+		{
+		    if (triggerPress (action, state, arguments))
 			return true;
+		}
 	    }
 	}
 
@@ -227,11 +269,11 @@ PrivateScreen::triggerButtonReleaseBindings (CompOption::Vector &options,
 
     foreach (CompOption &option, options)
     {
-	if (isTerminateBinding (option, type, state, &action))
+	if (isBound (option, type, state, &action))
 	{
 	    if (action->button ().button () == (int) event->button)
 	    {
-		if (action->terminate () (action, state, arguments))
+	        if (triggerRelease (action, state, arguments))
 		    return true;
 	    }
 	}
@@ -263,7 +305,7 @@ PrivateScreen::triggerKeyPressBindings (CompOption::Vector &options,
 	    {
 		if (!o.value ().action ().terminate ().empty ())
 		    o.value ().action ().terminate () (&o.value ().action (),
-						       state, noOptions);
+						       state, noOptions ());
 	    }
 	}
 
@@ -274,24 +316,19 @@ PrivateScreen::triggerKeyPressBindings (CompOption::Vector &options,
     state = CompAction::StateInitKey;
     foreach (CompOption &option, options)
     {
-	if (isInitiateBinding (option, CompAction::BindingTypeKey,
-			       state, &action))
+	if (isBound (option, CompAction::BindingTypeKey, state, &action))
 	{
 	    bindMods = modHandler->virtualToRealModMask (
 		action->key ().modifiers ());
 
+	    bool match = false;
 	    if (action->key ().keycode () == (int) event->keycode)
-	    {
-		if ((bindMods & modMask) == (event->state & modMask))
-		    if (action->initiate () (action, state, arguments))
-			return true;
-	    }
+		match = ((bindMods & modMask) == (event->state & modMask));
 	    else if (!xkbEvent && action->key ().keycode () == 0)
-	    {
-		if (bindMods == (event->state & modMask))
-		    if (action->initiate () (action, state, arguments))
-			return true;
-	    }
+		match = (bindMods == (event->state & modMask));
+
+	    if (match && triggerPress (action, state, arguments))
+		return true;
 	}
     }
 
@@ -316,25 +353,19 @@ PrivateScreen::triggerKeyReleaseBindings (CompOption::Vector &options,
 
     foreach (CompOption &option, options)
     {
-	if (isTerminateBinding (option, CompAction::BindingTypeKey,
-				state, &action))
+	if (isBound (option, CompAction::BindingTypeKey, state, &action))
 	{
 	    bindMods = modHandler->virtualToRealModMask (action->key ().modifiers ());
 
+	    bool match = false;
 	    if ((bindMods & modMask) == 0)
-	    {
-		if ((unsigned int) action->key ().keycode () ==
-						  (unsigned int) event->keycode)
-		{
-		    if (action->terminate () (action, state, arguments))
-			return true;
-		}
-	    }
+		match = ((unsigned int) action->key ().keycode () ==
+		         (unsigned int) event->keycode);
 	    else if (!xkbEvent && ((mods & modMask & bindMods) != bindMods))
-	    {
-		if (action->terminate () (action, state, arguments))
-		    return true;
-	    }
+	        match = true;
+
+	    if (match && triggerRelease (action, state, arguments))
+		return true;
 	}
     }
 
@@ -358,37 +389,39 @@ PrivateScreen::triggerStateNotifyBindings (CompOption::Vector  &options,
 
 	foreach (CompOption &option, options)
 	{
-	    if (isInitiateBinding (option, CompAction::BindingTypeKey,
-				   state, &action))
+	    if (isBound (option, CompAction::BindingTypeKey, state, &action))
 	    {
 		if (action->key ().keycode () == 0)
 		{
 		    bindMods =
 			modHandler->virtualToRealModMask (action->key ().modifiers ());
 
-		    if ((event->mods & modMask & bindMods) == bindMods)
+		    if ((event->mods & modMask) == bindMods)
 		    {
-			if (action->initiate () (action, state, arguments))
+		        if (triggerPress (action, state, arguments))
 			    return true;
 		    }
 		}
 	    }
 	}
     }
-    else
+    else if (event->event_type == KeyRelease)
     {
 	state = CompAction::StateTermKey;
 
 	foreach (CompOption &option, options)
 	{
-	    if (isTerminateBinding (option, CompAction::BindingTypeKey,
-				    state, &action))
+	    if (isBound (option, CompAction::BindingTypeKey, state, &action))
 	    {
-		bindMods = modHandler->virtualToRealModMask (action->key ().modifiers ());
+		bindMods = modHandler->virtualToRealModMask (
+		    action->key ().modifiers ());
+		unsigned int modKey =
+		    modHandler->keycodeToModifiers (event->keycode);
 
-		if ((event->mods & modMask & bindMods) != bindMods)
+		if ((event->mods && ((event->mods & modMask) != bindMods)) ||
+		    (!event->mods && (modKey == bindMods)))
 		{
-		    if (action->terminate () (action, state, arguments))
+		    if (triggerRelease (action, state, arguments))
 			return true;
 		}
 	    }
@@ -617,7 +650,7 @@ bool
 PrivateScreen::handleActionEvent (XEvent *event)
 {
     static CompOption::Vector o (8);
-    o.resize (8);
+    Window xid;
 
     o[0].setName ("event_window", CompOption::TypeInt);
     o[1].setName ("window", CompOption::TypeInt);
@@ -625,11 +658,26 @@ PrivateScreen::handleActionEvent (XEvent *event)
     o[3].setName ("x", CompOption::TypeInt);
     o[4].setName ("y", CompOption::TypeInt);
     o[5].setName ("root", CompOption::TypeInt);
+    o[6].reset ();
+    o[7].reset ();
 
     switch (event->type) {
     case ButtonPress:
+	/* We need to determine if we clicked on a parent frame
+	 * window, if so, pass the appropriate child window as
+	 * "window" and the frame as "event_window"
+	 */
+
+	xid = event->xbutton.window;
+
+	foreach (CompWindow *w, screen->windows ())
+	{
+	    if (w->priv->frame == xid)
+		xid = w->id ();
+	}
+
 	o[0].value ().set ((int) event->xbutton.window);
-	o[1].value ().set ((int) event->xbutton.window);
+	o[1].value ().set ((int) xid);
 	o[2].value ().set ((int) event->xbutton.state);
 	o[3].value ().set ((int) event->xbutton.x_root);
 	o[4].value ().set ((int) event->xbutton.y_root);
@@ -641,6 +689,7 @@ PrivateScreen::handleActionEvent (XEvent *event)
 	o[6].value ().set ((int) event->xbutton.button);
 	o[7].value ().set ((int) event->xbutton.time);
 
+	possibleTap = NULL;
 	foreach (CompPlugin *p, CompPlugin::getPlugins ())
 	{
 	    CompOption::Vector &options = p->vTable->getOptions ();
@@ -683,6 +732,7 @@ PrivateScreen::handleActionEvent (XEvent *event)
 	o[6].value ().set ((int) event->xkey.keycode);
 	o[7].value ().set ((int) event->xkey.time);
 
+	possibleTap = NULL;
 	foreach (CompPlugin *p, CompPlugin::getPlugins ())
 	{
 	    CompOption::Vector &options = p->vTable->getOptions ();
@@ -750,8 +800,6 @@ PrivateScreen::handleActionEvent (XEvent *event)
 
 		o[6].setName ("time", CompOption::TypeInt);
 		o[6].value ().set ((int) event->xcrossing.time);
-		
-		o.resize (7);
 
 		foreach (CompPlugin *p, CompPlugin::getPlugins ())
 		{
@@ -787,8 +835,6 @@ PrivateScreen::handleActionEvent (XEvent *event)
 
 		o[6].setName ("time", CompOption::TypeInt);
 		o[6].value ().set ((int) event->xcrossing.time);
-		
-		o.resize (7);
 
 		if (triggerEdgeEnter (edge, state, o))
 		    return true;
@@ -835,8 +881,6 @@ PrivateScreen::handleActionEvent (XEvent *event)
 		o[3].value ().set ((int) 0); /* fixme */
 		o[4].value ().set ((int) 0); /* fixme */
 		o[5].value ().set ((int) root);
-		
-		o.resize (6);
 
 		foreach (CompPlugin *p, CompPlugin::getPlugins ())
 		{
@@ -881,8 +925,6 @@ PrivateScreen::handleActionEvent (XEvent *event)
 		o[3].value ().set ((int) event->xclient.data.l[2] >> 16);
 		o[4].value ().set ((int) event->xclient.data.l[2] & 0xffff);
 		o[5].value ().set ((int) root);
-		
-		o.resize (6);
 
 		if (triggerEdgeEnter (edge, state, o))
 		    return true;
@@ -904,10 +946,15 @@ PrivateScreen::handleActionEvent (XEvent *event)
 		o[1].value ().set ((int) activeWindow);
 		o[2].value ().set ((int) stateEvent->mods);
 
-		o[3] = CompOption ("time", CompOption::TypeInt);
+		o[3].setName ("time", CompOption::TypeInt);
 		o[3].value ().set ((int) xkbEvent->time);
+		o[4].reset ();
+		o[5].reset ();
+		o[6].reset ();
+		o[7].value ().set ((int) xkbEvent->time);
 
-		o.resize (4);
+		if (stateEvent->event_type == KeyPress)
+		    possibleTap = NULL;
 
 		foreach (CompPlugin *p, CompPlugin::getPlugins ())
 		{
@@ -921,10 +968,12 @@ PrivateScreen::handleActionEvent (XEvent *event)
 		o[0].value ().set ((int) activeWindow);
 		o[1].value ().set ((int) activeWindow);
 
-		o[2] = CompOption ("time", CompOption::TypeInt);
+		o[2].setName ("time", CompOption::TypeInt);
 		o[2].value ().set ((int) xkbEvent->time);
+		o[3].reset ();
+		o[4].reset ();
+		o[5].reset ();
 
-		o.resize (3);
 
 		foreach (CompPlugin *p, CompPlugin::getPlugins ())
 		{
@@ -940,18 +989,96 @@ PrivateScreen::handleActionEvent (XEvent *event)
     return false;
 }
 
+void
+PrivateScreen::setDefaultWindowAttributes (XWindowAttributes *wa)
+{
+    wa->x		      = 0;
+    wa->y		      = 0;
+    wa->width		      = 1;
+    wa->height		      = 1;
+    wa->border_width	      = 0;
+    wa->depth		      = 0;
+    wa->visual		      = NULL;
+    wa->root		      = root;
+    wa->c_class		      = InputOnly;
+    wa->bit_gravity	      = NorthWestGravity;
+    wa->win_gravity	      = NorthWestGravity;
+    wa->backing_store	      = NotUseful;
+    wa->backing_planes	      = 0;
+    wa->backing_pixel	      = 0;
+    wa->save_under	      = false;
+    wa->colormap	      = None;
+    wa->map_installed	      = false;
+    wa->map_state	      = IsUnviewable;
+    wa->all_event_masks	      = 0;
+    wa->your_event_mask	      = 0;
+    wa->do_not_propagate_mask = 0;
+    wa->override_redirect     = true;
+    wa->screen		      = ScreenOfDisplay (dpy, screenNum);
+}
+
 void
 CompScreen::handleCompizEvent (const char         *plugin,
 			       const char         *event,
 			       CompOption::Vector &options)
-    WRAPABLE_HND_FUNC (7, handleCompizEvent, plugin, event, options)
+{
+    WRAPABLE_HND_FUNCTN (handleCompizEvent, plugin, event, options);
+    _handleCompizEvent (plugin, event, options);
+}
+
+void
+CompScreenImpl::_handleCompizEvent (const char         *plugin,
+			       const char         *event,
+			       CompOption::Vector &options)
+{
+}
 
 void
 CompScreen::handleEvent (XEvent *event)
 {
-    WRAPABLE_HND_FUNC (6, handleEvent, event)
+    WRAPABLE_HND_FUNCTN (handleEvent, event)
+    _handleEvent (event);
+}
 
-    CompWindow *w;
+void
+CompScreenImpl::alwaysHandleEvent (XEvent *event)
+{
+    /*
+     * Critical event handling that cannot be overridden by plugins
+     */
+    
+    if (event->type == ButtonPress || event->type == KeyPress)
+	priv->possibleTap = NULL;
+
+    eventHandled = true;  // if we return inside WRAPABLE_HND_FUNCTN
+
+    handleEvent (event);
+
+    bool keyEvent = (event->type == KeyPress || event->type == KeyRelease);
+
+    /* Always either replay the keyboard or consume the key
+     * event on keypresses */
+    if (keyEvent)
+    {
+	int mode = eventHandled ? AsyncKeyboard : ReplayKeyboard;
+	XAllowEvents (priv->dpy, mode, event->xkey.time);
+    }
+
+    if (priv->grabsEmpty () && event->type == KeyPress)
+    {
+	XUngrabKeyboard (priv->dpy, event->xkey.time);
+    }
+}
+
+void
+CompScreenImpl::_handleEvent (XEvent *event)
+{
+    /*
+     * Non-critical event handling that might be overridden by plugins
+     */
+
+    CompWindow *w = NULL;
+    XWindowAttributes wa;
 
     switch (event->type) {
     case ButtonPress:
@@ -970,15 +1097,16 @@ CompScreen::handleEvent (XEvent *event)
 	w = findWindow (priv->activeWindow);
 	if (w)
 	    priv->setCurrentOutput (w->outputDevice ());
+	break;
     default:
 	break;
     }
 
-    if (priv->handleActionEvent (event))
+    eventHandled = priv->handleActionEvent (event);
+    if (eventHandled)
     {
-	if (priv->grabs.empty ())
+	if (priv->grabsEmpty ())
 	    XAllowEvents (priv->dpy, AsyncPointer, event->xbutton.time);
-
 	return;
     }
 
@@ -991,7 +1119,8 @@ CompScreen::handleEvent (XEvent *event)
 	break;
     case ConfigureNotify:
 	w = findWindow (event->xconfigure.window);
-	if (w && !w->frame ())
+
+	if (w && !w->priv->frame)
 	{
 	    w->priv->configure (&event->xconfigure);
 	}
@@ -999,10 +1128,8 @@ CompScreen::handleEvent (XEvent *event)
 	{
 	    w = findTopLevelWindow (event->xconfigure.window);
 
-	    if (w && w->frame () == event->xconfigure.window)
-	    {
+	    if (w && w->priv->frame == event->xconfigure.window)
 		w->priv->configureFrame (&event->xconfigure);
-	    }
 	    else
 	    {
 		if (event->xconfigure.window == priv->root)
@@ -1011,16 +1138,91 @@ CompScreen::handleEvent (XEvent *event)
 	}
 	break;
     case CreateNotify:
-	w = findTopLevelWindow (event->xcreatewindow.window, true);
+    {
+	bool create = true;
 
-        if (event->xcreatewindow.parent == priv->root &&
-	    (!w || w->frame () != event->xcreatewindow.window))
+	/* Failure means that window has been destroyed. We still have to add 
+	 * the window to the window list as we might get configure requests
+	 * which require us to stack other windows relative to it. Setting
+	 * some default values if this is the case. */
+	if (!XGetWindowAttributes (priv->dpy, event->xcreatewindow.window, &wa))
+	    priv->setDefaultWindowAttributes (&wa);
+
+	foreach (CompWindow *w, screen->windows ())
 	{
-	    new CompWindow (event->xcreatewindow.window, priv->getTopWindow ());
+	    if (w->priv->serverFrame == event->xcreatewindow.window)
+	    {
+		w->priv->frame = event->xcreatewindow.window;
+		w->priv->updatePassiveButtonGrabs ();
+		create = false;
+	    }
 	}
+
+	foreach (CompWindow *w, priv->destroyedWindows)
+	{
+	    if (w->priv->serverId == event->xcreatewindow.window)
+	    {
+		/* Previously destroyed window
+		 * plugins were keeping around
+		 * in order to avoid an xid conflict,
+		 * destroy it right away and manage
+		 * the new window */
+
+		StackDebugger *dbg = StackDebugger::Default ();
+
+		while (w->priv->destroyRefCnt)
+		    w->destroy ();
+
+		if (dbg)
+		    dbg->removeDestroyedFrame (event->xcreatewindow.window);
+	    }
+
+	}
+
+	if (wa.root != event->xcreatewindow.parent)
+	    create = false;
+
+	if (create)
+	{
+	    /* Track the window if it was created on this
+	     * screen, otherwise we still need to register
+	     * for FocusChangeMask. Also, we don't want to
+	     * manage it straight away - in reality we want
+	     * that to wait until the map request */
+	    if ((wa.root == priv->root))
+	    {
+		PrivateWindow::createCompWindow (priv->getTopWindow (), wa, event->xcreatewindow.window);
+            }
+	    else
+		XSelectInput (priv->dpy, event->xcreatewindow.window,
+			      FocusChangeMask);
+	}
+	else
+	    compLogMessage ("core", CompLogLevelDebug, "refusing to manage window 0x%x", (unsigned int) event->xcreatewindow.window);
+
 	break;
+    }
     case DestroyNotify:
 	w = findWindow (event->xdestroywindow.window);
+
+	/* It is possible that some plugin might call
+	 * w->destroy () before the window actually receives
+	 * its first DestroyNotify event which would mean
+	 * that it is already in the list of destroyed
+	 * windows, so check that list too */
+
+	if (!w)
+	{
+	    foreach (CompWindow *dw, priv->destroyedWindows)
+	    {
+		if (dw->priv->serverId == event->xdestroywindow.window)
+		{
+		    w = dw;
+		    break;
+		}
+	    }
+	}
+
 	if (w)
 	{
 	    w->moveInputFocusToOtherWindow ();
@@ -1028,13 +1230,23 @@ CompScreen::handleEvent (XEvent *event)
 	}
 	break;
     case MapNotify:
-	w = findWindow (event->xmap.window);
+
+	/* Search in already-created windows for this window */
+	if (!w)
+	    w = findWindow (event->xmap.window);
+
 	if (w)
 	{
 	    if (w->priv->pendingMaps)
 	    {
-		if (!w->priv->frame)
+		/* The only case where this happens
+		 * is where the window unmaps itself
+		 * but doesn't get destroyed so when
+		 * it re-maps we need to reparent it */
+
+		if (!w->priv->serverFrame)
 		    w->priv->reparent ();
+
 		w->priv->managed = true;
 	    }
 
@@ -1047,6 +1259,7 @@ CompScreen::handleEvent (XEvent *event)
 
 	    w->map ();
 	}
+
 	break;
     case UnmapNotify:
 	w = findWindow (event->xunmap.window);
@@ -1060,55 +1273,100 @@ CompScreen::handleEvent (XEvent *event)
 	    }
 	    else /* X -> Withdrawn */
 	    {
-		/* Iconic -> Withdrawn */
+		/* Iconic -> Withdrawn:
+		 *
+		 * The window is already unmapped so we need to check the
+		 * synthetic UnmapNotify that comes and withdraw the window here */
 		if (w->state () & CompWindowStateHiddenMask)
 		{
 		    w->priv->minimized = false;
-
 		    w->changeState (w->state () & ~CompWindowStateHiddenMask);
 
 		    priv->updateClientList ();
+		    w->priv->withdraw ();
 		}
-		else /* Closing */
-		    w->windowNotify (CompWindowNotifyClose);
-
-		if (!w->overrideRedirect ())
-		    priv->setWmState (WithdrawnState, w->id ());
-
-		w->priv->placed = false;
-		w->priv->managed = false;
-		if (w->priv->frame)
+		/* Closing:
+		 *
+		 * ICCCM Section 4.1.4 says that clients need to send
+		 * a synthetic UnmapNotify for every real unmap
+		 * in order to reflect the change in state, but
+		 * since we already withdraw the window on the real
+		 * UnmapNotify, no need to do it again on the synthetic
+		 * one. */
+		else if (!event->xunmap.send_event)
 		{
-		    w->priv->unreparent ();
+		    w->windowNotify (CompWindowNotifyClose);
+		    w->priv->withdraw ();
 		}
 	    }
 
-	    w->unmap ();
+	    if (!event->xunmap.send_event)
+	    {
+		w->unmap ();
 
-	    if (!w->shaded () && !w->priv->pendingMaps)
-		w->moveInputFocusToOtherWindow ();
+		if (!w->shaded () && !w->priv->pendingMaps)
+		    w->moveInputFocusToOtherWindow ();
+	    }
 	}
 	break;
     case ReparentNotify:
+
 	w = findWindow (event->xreparent.window);
-	if (!w && event->xreparent.parent == priv->root)
-	{
-	    new CompWindow (event->xreparent.window, priv->getTopWindow ());
-	}
-	else if (w && !(event->xreparent.parent == w->priv->wrapper ||
-		 event->xreparent.parent == priv->root))
-	{
-	    /* This is the only case where a window is removed but not
-	       destroyed. We must remove our event mask and all passive
-	       grabs. */
-	    XSelectInput (priv->dpy, w->id (), NoEventMask);
-	    XShapeSelectInput (priv->dpy, w->id (), NoEventMask);
-	    XUngrabButton (priv->dpy, AnyButton, AnyModifier, w->id ());
 
-	    w->moveInputFocusToOtherWindow ();
+	/* If this window isn't part of our tracked window
+	 * list and was reparented into the root window then
+	 * we need to track it */
+	if (!w)
+	{
+	    if (event->xreparent.parent == priv->root)
+	    {
+		/* Failure means that window has been destroyed. We still have to add 
+		 * the window to the window list as we might get configure requests
+		 * which require us to stack other windows relative to it. Setting
+		 * some default values if this is the case. */
+		if (!XGetWindowAttributes (priv->dpy, event->xcreatewindow.window, &wa))
+		    priv->setDefaultWindowAttributes (&wa);
 
-	    w->destroy ();
+		PrivateWindow::createCompWindow (priv->getTopWindow (), wa, event->xcreatewindow.window);
+		break;
+	    }
+	    else
+	    {
+		/* It is possible that some plugin might call
+		 * w->destroy () before the window actually receives
+		 * its first ReparentNotify event which would mean
+		 * that it is already in the list of destroyed
+		 * windows, so check that list too */
+
+		foreach (CompWindow *dw, priv->destroyedWindows)
+		{
+		    if (dw->priv->serverId == event->xreparent.window)
+		    {
+			w = dw;
+			break;
+		    }
+		}
+	    }
 	}
+
+	/* This is the only case where a window is removed but not
+	   destroyed. We must remove our event mask and all passive
+	   grabs. */
+
+        if (w)
+	{
+	    if ((w->priv->wrapper && event->xreparent.parent != w->priv->wrapper) ||
+		(!w->priv->wrapper && event->xreparent.parent != priv->root))
+	    {
+		w->moveInputFocusToOtherWindow ();
+		w->destroy ();
+
+		XSelectInput (priv->dpy, w->id (), NoEventMask);
+		XShapeSelectInput (priv->dpy, w->id (), NoEventMask);
+		XUngrabButton (priv->dpy, AnyButton, AnyModifier, w->id ());
+	    }
+        }
+
 	break;
     case CirculateNotify:
 	w = findWindow (event->xcirculate.window);
@@ -1116,28 +1374,28 @@ CompScreen::handleEvent (XEvent *event)
 	    w->priv->circulate (&event->xcirculate);
 	break;
     case ButtonPress:
-	if (event->xbutton.root == priv->root)
+	if (event->xbutton.button == Button1 ||
+	    event->xbutton.button == Button2 ||
+	    event->xbutton.button == Button3)
 	{
-	    if (event->xbutton.button == Button1 ||
-		event->xbutton.button == Button2 ||
-		event->xbutton.button == Button3)
+	    w = findTopLevelWindow (event->xbutton.window);
+	    if (w)
 	    {
-		w = findTopLevelWindow (event->xbutton.window);
-		if (w)
+		if (priv->optionGetRaiseOnClick ())
 		{
-		    if (priv->optionGetRaiseOnClick ())
-			w->updateAttributes (
-					CompStackingUpdateModeAboveFullscreen);
-
-		    if (w->id () != priv->activeWindow)
-			if (!(w->type () & CompWindowTypeDockMask))
-			    w->moveInputFocusTo ();
+		    w->updateAttributes (CompStackingUpdateModeAboveFullscreen);
 		}
-	    }
 
-	    if (priv->grabs.empty ())
-		XAllowEvents (priv->dpy, ReplayPointer, event->xbutton.time);
+	        if (w->id () != priv->activeWindow)
+		    if (!(w->type () & CompWindowTypeDockMask))
+			if (w->focus ())
+			    w->moveInputFocusTo ();
+	    }
 	}
+
+	if (priv->grabsEmpty ())
+	    XAllowEvents (priv->dpy, ReplayPointer, event->xbutton.time);
+
 	break;
     case PropertyNotify:
 	if (event->xproperty.atom == Atoms::winType)
@@ -1276,7 +1534,7 @@ CompScreen::handleEvent (XEvent *event)
     case ClientMessage:
 	if (event->xclient.message_type == Atoms::winActive)
 	{
-	    w = findWindow (event->xclient.window);
+	    w = findTopLevelWindow (event->xclient.window);
 	    if (w)
 	    {
 		/* use focus stealing prevention if request came
@@ -1389,37 +1647,39 @@ CompScreen::handleEvent (XEvent *event)
 		unsigned int   xwcm = 0;
 		XWindowChanges xwc;
 		int            gravity;
+		int	       value_mask;
 		unsigned int   source;
 
+		gravity = (event->xclient.data.l[0] & 0xFF);		
+		value_mask = (event->xclient.data.l[0] & 0xF00) >> 8;
+		source = (event->xclient.data.l[0] & 0xF000) >> 12;
+
 		memset (&xwc, 0, sizeof (xwc));
 
-		if (event->xclient.data.l[0] & (1 << 8))
+		if (value_mask & CWX)
 		{
 		    xwcm |= CWX;
 		    xwc.x = event->xclient.data.l[1];
 		}
 
-		if (event->xclient.data.l[0] & (1 << 9))
+		if (value_mask & CWY)
 		{
 		    xwcm |= CWY;
 		    xwc.y = event->xclient.data.l[2];
 		}
 
-		if (event->xclient.data.l[0] & (1 << 10))
+		if (value_mask & CWWidth)
 		{
 		    xwcm |= CWWidth;
 		    xwc.width = event->xclient.data.l[3];
 		}
 
-		if (event->xclient.data.l[0] & (1 << 11))
+		if (value_mask & CWHeight)
 		{
 		    xwcm |= CWHeight;
 		    xwc.height = event->xclient.data.l[4];
 		}
 
-		gravity = event->xclient.data.l[0] & 0xFF;
-		source  = (event->xclient.data.l[0] >> 12) & 0xF;
-
 		w->moveResize (&xwc, xwcm, gravity, source);
 	    }
 	}
@@ -1462,7 +1722,9 @@ CompScreen::handleEvent (XEvent *event)
 			w->minimize ();
 		}
 		else if (event->xclient.data.l[0] == NormalState)
+		{
 		    w->unminimize ();
+		}
 	    }
 	}
 	else if (event->xclient.message_type == Atoms::showingDesktop)
@@ -1518,7 +1780,8 @@ CompScreen::handleEvent (XEvent *event)
 	modHandler->updateModifierMappings ();
 	break;
     case MapRequest:
-	w = findWindow (event->xmaprequest.window);
+	w = screen->findWindow (event->xmaprequest.window);
+
 	if (w)
 	{
 	    XWindowAttributes attr;
@@ -1617,7 +1880,13 @@ CompScreen::handleEvent (XEvent *event)
 	    xwc.border_width = event->xconfigurerequest.border_width;
 
 	    if (w)
+	    {
+		/* Any window that receives a ConfigureRequest
+		 * is not override redirect, and may have changed
+		 * to being not override redirect */
+		w->priv->setOverrideRedirect (false);
 		w->configureXWindow (xwcm, &xwc);
+	    }
 	    else
 		XConfigureWindow (priv->dpy, event->xconfigurerequest.window,
 				  xwcm, &xwc);
@@ -1626,32 +1895,179 @@ CompScreen::handleEvent (XEvent *event)
     case CirculateRequest:
 	break;
     case FocusIn:
-	if (event->xfocus.mode != NotifyGrab)
+    {
+	if (!XGetWindowAttributes (priv->dpy, event->xfocus.window, &wa))
+	    priv->setDefaultWindowAttributes (&wa);
+
+        /* If the call to XGetWindowAttributes failed it means
+         * the window was destroyed, so track the focus change
+         * anyways since we need to increment activeNum
+         * and the passive button grabs and then we will
+         * get the DestroyNotify later and change the focus
+         * there
+	 */
+
+	if (wa.root == priv->root)
 	{
-	    w = findTopLevelWindow (event->xfocus.window);
-	    if (w && w->managed ())
+	    if (event->xfocus.mode == NotifyGrab)
+		priv->grabNotified ();
+	    else if (event->xfocus.mode == NotifyUngrab)
+		priv->ungrabNotified ();
+	    else
 	    {
-		unsigned int state = w->state ();
+		CompWindowList dockWindows;
+		XWindowChanges xwc;
+		unsigned int   mask;
 
-		if (w->id () != priv->activeWindow)
+		w = findTopLevelWindow (event->xfocus.window);
+		if (w && w->managed ())
 		{
-		    w->windowNotify (CompWindowNotifyFocusChange);
+		    unsigned int state = w->state ();
 
-		    priv->activeWindow = w->id ();
-		    w->priv->activeNum = priv->activeNum++;
+		    if (priv->nextActiveWindow == event->xfocus.window)
+			priv->nextActiveWindow = None;
 
-		    priv->addToCurrentActiveWindowHistory (w->id ());
+		    if (w->id () != priv->activeWindow)
+		    {
+			CompWindow     *active = screen->findWindow (priv->activeWindow);
 
-		    XChangeProperty (priv->dpy , priv->root,
-				     Atoms::winActive,
-				     XA_WINDOW, 32, PropModeReplace,
-				     (unsigned char *) &priv->activeWindow, 1);
+			priv->activeWindow = w->id ();
+			w->priv->activeNum = priv->nextActiveNum();
+
+			if (active)
+			{
+			    CompWindowList windowsLostFocus;
+			    /* If this window lost focus and was above a fullscreen window
+			     * and is no longer capable of being focused (eg, it is
+			     * not visible on this viewport) then we need to check if
+			     * any other windows below it are also now no longer capable
+			     * of being focused and restack them in the highest position
+			     * below docks that they are allowed to take */
+			    if (!active->focus ())
+			    {
+				windowsLostFocus.push_back (active);
+				for (CompWindow *fsw = active->prev; fsw; fsw = fsw->prev)
+				{
+				    if (!fsw->focus () &&
+					fsw->managed () &&
+					!(fsw->type () & (CompWindowTypeDockMask |
+							  CompWindowTypeFullscreenMask)) &&
+					!fsw->overrideRedirect ())
+					windowsLostFocus.push_back (fsw);
+
+				    if (fsw->type () & CompWindowTypeFullscreenMask)
+				    {
+					/* This will be the window that we must lower relative to */
+					CompWindow *sibling = PrivateWindow::findValidStackSiblingBelow (active, fsw);
+
+					if (sibling)
+					{
+					    for (CompWindowList::reverse_iterator rit = windowsLostFocus.rbegin ();
+						 rit != windowsLostFocus.rend (); rit++)
+					    {
+						(*rit)->restackAbove (sibling);
+					    }
+					}
+
+					break;
+				    }
+				}
+			    }
+
+			    active->changeState (active->focused () ?
+						 active->state () | CompWindowStateFocusedMask :
+						 active->state () & ~CompWindowStateFocusedMask);
+
+			    active->priv->updatePassiveButtonGrabs ();
+			}
+
+			if (w->focused ())
+			    state |= w->state () | CompWindowStateFocusedMask;
+			else
+			    state &= w->state () & ~CompWindowStateFocusedMask;
+
+			w->priv->updatePassiveButtonGrabs ();
+
+			priv->addToCurrentActiveWindowHistory (w->id ());
+
+			XChangeProperty (priv->dpy , priv->root,
+					 Atoms::winActive,
+					 XA_WINDOW, 32, PropModeReplace,
+					 (unsigned char *) &priv->activeWindow, 1);
+
+			w->windowNotify (CompWindowNotifyFocusChange);
+		    }
+
+		    state &= ~CompWindowStateDemandsAttentionMask;
+		    w->changeState (state);
+	        }
+		else if (event->xfocus.window == priv->root)
+		{
+		    /* Don't ever let the focus go to the root
+		     * window except in grab cases
+		     *
+		     * FIXME: There might be a case where we have to
+		     * handle root windows of other screens here, but
+		     * the other window managers should handle that
+		     */
+
+		    if (event->xfocus.detail == NotifyDetailNone ||
+			(event->xfocus.mode == NotifyNormal &&
+			 event->xfocus.detail == NotifyInferior))
+		    {
+			priv->activeWindow = None;
+
+			if (event->xfocus.detail == NotifyDetailNone ||
+			    (event->xfocus.mode == NotifyNormal &&
+			     event->xfocus.detail == NotifyInferior))
+			{
+			    screen->focusDefaultWindow ();
+			}
+		    }
+		}
+
+		/* Ensure that docks are stacked in the right place
+		 *
+		 * When a normal window gets the focus and is above a
+		 * fullscreen window, restack the docks to be above
+		 * the highest level mapped and visible normal window,
+		 * otherwise put them above the highest fullscreen window
+		 */
+		if (w)
+		{
+		    if (PrivateWindow::stackDocks (w, dockWindows, &xwc, &mask))
+		    {
+			Window sibling = xwc.sibling;
+			xwc.stack_mode = Above;
+
+			/* Then update the dock windows */
+			foreach (CompWindow *dw, dockWindows)
+			{
+			    xwc.sibling = sibling;
+			    dw->configureXWindow (mask, &xwc);
+			}
+		    }
 		}
 
-		state &= ~CompWindowStateDemandsAttentionMask;
-		w->changeState (state);
 	    }
 	}
+	else
+	{
+	    CompWindow *w;
+
+	    w = screen->findWindow (priv->activeWindow);
+
+	    priv->nextActiveWindow = None;
+	    priv->activeWindow = None;
+
+	    if (w)
+		w->priv->updatePassiveButtonGrabs ();
+	}
+    }
+    break;
+    case FocusOut:
+	if (event->xfocus.mode == NotifyUngrab)
+	    priv->ungrabNotified ();
 	break;
     case EnterNotify:
 	if (event->xcrossing.root == priv->root)
@@ -1664,7 +2080,7 @@ CompScreen::handleEvent (XEvent *event)
 	    priv->below = w->id ();
 
 	    if (!priv->optionGetClickToFocus () &&
-		priv->grabs.empty ()                                 &&
+		priv->grabsEmpty ()                                 &&
 		event->xcrossing.mode   != NotifyGrab                &&
 		event->xcrossing.detail != NotifyInferior)
 	    {
